from ... import logger

def args_sink(parser):
    pass

class Sink:
    def __init__(self, ctrl):
        self.log = logger.new(f"tkGUI.ctrl.{type(self).__name__}")
        self.ctrl = ctrl

    def start(self):
        pass
    def stop(self):
        pass
    def reset(self):
        pass

    def draw_tb(self):
        pass
    def draw_cl(self):
        pass



"""Create a CUI view"""

from ..base import View as _View

class View(_View):
    """Parent CUI view class"""
    def __init__(self, **kwargs):
        self.running = True

    def mainloop(self):
        while self.running:
            try:
                self.menu()
            except KeyboardInterrupt:
                pass


    def menu(self):
        pass



"""Controller for FILE sink"""
import argparse
import datetime as dt
import tkinter as tk

import numpy as np

from ...utils import args
from .sink import Sink, args_sink

from .dispatch import CMD
# from .arc.plot_base import define_args as freq_args
from .panels import PanelController, PanelChild, Panel
from .plot_base import FreqPlotController, BlitPlot

from .swept import plots

from ...utils import dialog
from ...utils.time import strfmt_td
from ...model.sink.file import Format

class SinkConfig:
    pass

def args_file(parser: argparse.ArgumentParser):
    sink = args.get_group(parser, "Sink (FILE)")
    args_sink(sink)

class SinkFile(Sink):
    def __init__(self, ctrl):
        super().__init__(ctrl)
        self.ctrl.view.sink.tb_sld_samp.scale.config(from_=0, to=self.ctrl.model.sink.max_samp) # resolution=self.model.block_size
        self.ctrl.view.sink.tb_sld_samp.scale.config(command=self.handle_sld_samp)

        self.ctrl.view.sink.tb_btn_prev.config(command=self.prev)
        self.ctrl.view.sink.tb_btn_next.config(command=self.next)

        self.ctrl.view.sink.cl_btn_file.config(command=self.handle_btn_file)
        self.ctrl.view.sink.cl_cb_file_fmt.config(values=list([v.name for v in Format]))
        self.ctrl.view.sink.cl_cb_file_fmt.bind("<<ComboboxSelected>>", self.handle_event)

    def start(self):
        if self.ctrl.model.sink.get_path() is None:
            print("sink is not fully initialized, failed to get filepath")
            return
        self.ctrl.view.tb_btn_start.config(state=tk.DISABLED)
        self.ctrl.view.tb_btn_stop.config(state=tk.ACTIVE)
        self.ctrl.dispatch.send(CMD.START)

    def stop(self):
        self.ctrl.view.tb_btn_stop.config(state=tk.DISABLED)
        self.ctrl.view.tb_btn_start.config(state=tk.ACTIVE)
        self.ctrl.dispatch.send(CMD.STOP)

    def reset(self):
        self.ctrl.dispatch.send(CMD.RESET)
        self.stop()

    def prev(self):
        self.ctrl.dispatch.send(CMD.PREV)

    def next(self):
        self.ctrl.dispatch.send(CMD.NEXT)

    def draw_tb(self):
        self.ctrl.view.sink.tb_var_samp.set(self.ctrl.model.sink.cur_samp)

        self.ctrl.view.sink.tb_var_time_cur.set(strfmt_td(dt.timedelta(seconds=self.ctrl.model.sink.cur_time())))
        self.ctrl.view.sink.tb_var_time_tot.set(strfmt_td(dt.timedelta(seconds=self.ctrl.model.sink.tot_time())))

    def draw_cl(self):
        self.ctrl.view.sink.cl_var_file.set(str(self.ctrl.model.sink.get_path()))
        self.ctrl.view.sink.cl_var_file_fmt.set(str(self.ctrl.model.sink.get_fmt().name))


    # --- GUI bind events and setters --- #
    def handle_event(self, event):
        if event.widget == self.ctrl.view.sink.cl_cb_file_fmt:
            self.set_dtype(self.ctrl.view.sink.cl_var_file_fmt.get())
    def handle_btn_file(self):
        self.log.trace("handle_btn_file()")
        self.set_path(dialog.get_file(False))

    def handle_sld_samp(self, *args):
        self.log.trace("handle_sld_samp()")
        self.set_samp(self.ctrl.view.sink.tb_var_samp.get())

    def set_samp(self, samp):
        self.log.trace("set_samp(%s)", samp)
        self.stop()
        self.ctrl.model.sink.cur_samp = samp
        self.draw_tb()
        # print(samp)

    def set_path(self, path):
        self.log.trace("set_path(%s)", path)
        self.ctrl.model.sink.set_path(path)
        print(f"Path set to '{self.ctrl.model.sink.get_path()}'")
        self.ctrl.view.sink.tb_sld_samp.scale.config(from_=0, to=self.ctrl.model.sink.max_samp) # resolution=self.model.block_size
        self.draw_tb()
        self.draw_cl()
    def set_dtype(self, dtype):
        self.log.trace("set_dtype(%s)", dtype)
        self.ctrl.model.sink.set_fmt(dtype)
        self.draw_tb()
        self.draw_cl()



import argparse
import sys

from .. import err
from ..specan import SpecAn

from ..config import config, View, Mode, Sink
from ..model.model import ModelArgs
from ..controller.controller import ControllerArgs

from ..utils import args

def define_args():
    parser = argparse.ArgumentParser("pyspecan", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    args.update_register(parser)
    # mon = parser.add_argument_group("developer toggles")
    mon = args.get_group(parser, "developer toggles")
    mon.add_argument("--mon_mem", action="store_true")
    mon.add_argument("--profile", action="store_true")
    return parser

def _main(args):
    SpecAn(**vars(args))

def _process_args(parser):
    run_help = False
    if "-h" in sys.argv:
        run_help = True
        sys.argv.pop(sys.argv.index("-h"))
    elif "--help" in sys.argv:
        run_help = True
        sys.argv.pop(sys.argv.index("--help"))
    args, remaining = parser.parse_known_args()

    view = View.get(args.view)
    if view == View.NONE:
        raise err.UnknownOption(f"Unknown view {args.view}")

    mode = Mode.get(args.mode)
    if mode == Mode.NONE:
        raise err.UnknownOption(f"Unknown mode {args.mode}")

    sink = Sink.get(args.sink)
    if sink == Sink.NONE:
        raise err.UnknownOption(f"Unknown sink: {args.sink}")

    if args.mon_mem:
        config.MON_MEM = True
    if args.profile:
        config.PROFILE = True

    ModelArgs(mode, sink, parser)
    ControllerArgs(view, mode, sink, parser)

    args = parser.parse_args()
    if run_help:
        parser.print_help()
        exit()
    return args

def main():
    parser = define_args()
    parser.add_argument("-v", "--view", type=str, default=View.tkGUI.name, choices=View.choices())
    parser.add_argument("-m", "--mode", type=str.upper, default=Mode.SWEPT.name, choices=Mode.choices())
    parser.add_argument("-s", "--sink", type=str.upper, default=Sink.FILE.name, choices=Sink.choices())
    _main(_process_args(parser))



import logging
from logging.handlers import RotatingFileHandler

LOGGING = False
LOG_LEVEL_TRACE = 5

logging.addLevelName(LOG_LEVEL_TRACE, "TRACE")

class TraceLogger(logging.Logger):
    def trace(self, msg, *args, **kwargs):
        if self.isEnabledFor(LOG_LEVEL_TRACE):
            self._log(LOG_LEVEL_TRACE, msg, args, **kwargs)

formatter = logging.Formatter(
    "%(asctime)s <%(name)s> %(levelname)s: %(message)s",
    datefmt="%Y%m%d_%H%M%S"
)
if LOGGING:
    handler_file = RotatingFileHandler("pyspecan.log", maxBytes=1024*1024, backupCount=3)
    handler_file.setLevel(logging.NOTSET)
    handler_file.setFormatter(formatter)

    handler_console = logging.StreamHandler()
    handler_console.setLevel(logging.INFO)
    handler_console.setFormatter(formatter)

def new(name: str, level = None, ch=True, fh=True):
    if level is None:
        level = logging.DEBUG
    logger = TraceLogger(f"pyspecan.{name}")

    if LOGGING and ch:
        logger.addHandler(handler_console)
    if LOGGING and fh:
        logger.addHandler(handler_file)
    return logger



from collections import Counter
import linecache
import os
import tracemalloc

import cProfile

from ..config import config

class Memory:
    __instance = None

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
            cls._peak = None
            cls._peak_tot = 0
        return cls.__instance

    def __str__(self):
        return f"{tracemalloc.get_traced_memory()}"

    def start(self):
        tracemalloc.start()

    def stop(self):
        tracemalloc.stop()

    def peak(self):
        if self._peak is None:
            _size_1, self._peak = tracemalloc.get_traced_memory()
            tracemalloc.reset_peak()
            self._peak_tot = self._peak_tot if self._peak < self._peak_tot else self._peak
        else:
            _size_2, peak_2 = tracemalloc.get_traced_memory()
            tracemalloc.reset_peak()
            print(f"Peak 1 {self._peak}, 2: {peak_2}, tot {self._peak_tot}")
            self._peak_tot = self._peak_tot if peak_2 < self._peak_tot else peak_2
            self._peak = None

    def _snapshot(self):
        return tracemalloc.take_snapshot()

    def show(self, snapshot=None, key_type='lineno', limit=3):
        if snapshot is None:
            snapshot = self._snapshot()
        snapshot = snapshot.filter_traces((
            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
            tracemalloc.Filter(False, "<frozen importlib._bootstrap_external>"),
            tracemalloc.Filter(False, "<unknown>"),
        ))
        top_stats = snapshot.statistics(key_type, cumulative=True)

        print("Top %s lines" % limit)
        for index, stat in enumerate(top_stats[:limit], 1):
            frame = stat.traceback[0]
            # replace "/path/to/module/file.py" with "module/file.py"
            filename = os.sep.join(frame.filename.split(os.sep)[-2:])
            print("#%s: %s:%s: %.1f KiB"
                % (index, filename, frame.lineno, stat.size / 1024))
            line = linecache.getline(frame.filename, frame.lineno).strip()
            if line:
                print('    %s' % line)

        other = top_stats[limit:]
        if other:
            size = sum(stat.size for stat in other)
            print("%s other: %.1f KiB" % (len(other), size / 1024))
        total = sum(stat.size for stat in top_stats)
        print("Total allocated size: %.1f KiB" % (total / 1024))

class Profile:
    __instance = None

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
            cls._pr = cProfile.Profile()
        return cls.__instance

    def enable(self):
        self._pr.enable()

    def disable(self):
        self._pr.disable()

    def show(self, sort=-1):
        self._pr.print_stats(sort)

    def dump(self, path):
        self._pr.dump_stats(path)



import matplotlib.pyplot as plt
import matplotlib.animation as manim

class FFmpegMP4:
    """
    with rp.viz.save.FFmpegMP4(fig, fps, "video.mp4") as writer:
        ...
        plt.plot()
        ...
        writer.grab_frame()

    writer = rp.viz.save.FFmpegMP4(fig, fps, "video.mp4")
    writer.open()
    ...
    writer.close()
    """
    def __init__(self, fig, fps, path, metadata=None, dpi=100):
        self.fig = fig
        self.path = path
        self.dpi = dpi
        if metadata is None:
            metadata = dict(title="Video", artist="Author")

        ImplWriter = manim.writers["ffmpeg"]
        self.writer = ImplWriter(fps=fps, metadata=metadata)
        self.writer.setup(fig, path)

    def open(self):
        self.wh = self.writer.saving(self.fig, self.path, self.dpi)

    def close(self):
        self.writer.finish()

    def __call__(self):
        return self.__enter__()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def grab_frame(self):
        return self.writer.grab_frame()



import tkinter as tk
import tkinter.ttk as ttk

from ...backend.tk import widgets

from ...utils import args
from .sink import Sink, args_sink

def args_file(parser):
    sink = args.get_group(parser, "Sink (FILE)")
    args_sink(sink)

class SinkFile(Sink):
    __slots__ = (
        # toolbar
        "tb_btn_prev", "tb_btn_next",
        "tb_var_samp", "tb_sld_samp",
        "tb_var_time_cur", "tb_lbl_time_cur",
        "tb_var_time_tot", "tb_lbl_time_tot",
        # control panel
        "cl_var_file", "cl_btn_file", "cl_ent_file",
        "cl_var_file_fmt", "cl_cb_file_fmt"
    )
    def __init__(self, view, **kwargs):
        super().__init__(view, **kwargs)
        # toolbar
        self.tb_btn_prev: ttk.Button = None # type: ignore
        self.tb_btn_next: ttk.Button = None # type: ignore
        self.tb_var_samp: tk.IntVar = None # type: ignore
        self.tb_sld_samp: widgets.Scale = None # type: ignore
        self.tb_var_time_cur: tk.StringVar = None # type: ignore
        self.tb_lbl_time_cur: ttk.Label = None # type: ignore
        self.tb_var_time_tot: tk.StringVar = None # type: ignore
        self.tb_lbl_time_tot: ttk.Label = None # type: ignore
        # control panel
        self.cl_var_file: tk.StringVar = None # type: ignore
        self.cl_btn_file: ttk.Button = None # type: ignore
        self.cl_ent_file: ttk.Entry = None # type: ignore
        self.cl_var_file_fmt: tk.StringVar = None # type: ignore
        self.cl_cb_file_fmt: ttk.Combobox = None # type: ignore

    def draw_tb(self, parent, col=0):
        self.tb_var_samp = tk.IntVar(parent)
        self.tb_sld_samp = widgets.Scale(
            parent, variable=self.tb_var_samp, length=150
        )
        self.tb_sld_samp.grid(row=0,rowspan=2,column=col, sticky=tk.NSEW)
        col += 1
        self.tb_var_time_cur = tk.StringVar(parent)
        self.tb_var_time_tot = tk.StringVar(parent)
        self.tb_lbl_time_cur = ttk.Label(parent, textvariable=self.tb_var_time_cur)
        self.tb_lbl_time_cur.grid(row=0,column=col)
        self.tb_lbl_time_tot = ttk.Label(parent, textvariable=self.tb_var_time_tot)
        self.tb_lbl_time_tot.grid(row=1,column=col)
        col += 1
        return col
    def draw_cl(self, parent, row=0):
        self.cl_var_file = tk.StringVar(parent)
        self.cl_btn_file = ttk.Button(parent, text="File")
        self.cl_btn_file.grid(row=row,column=0, sticky=tk.W)
        self.cl_ent_file = ttk.Entry(parent, textvariable=self.cl_var_file, state=tk.DISABLED, width=10)
        self.cl_ent_file.grid(row=row,column=1,columnspan=2, sticky=tk.NSEW)
        row += 1
        ttk.Label(parent, text="Format:").grid(row=row,column=0,sticky=tk.W)
        self.cl_var_file_fmt = tk.StringVar(parent)
        self.cl_cb_file_fmt = ttk.Combobox(parent, textvariable=self.cl_var_file_fmt, width=5)
        self.cl_cb_file_fmt.grid(row=row,column=1, sticky=tk.W)
        parent.columnconfigure(2, weight=1)
        return row

    def finish_tb(self):
        row, col = self.view.tb_fr_btn.grid_size()
        col += 1
        self.tb_btn_prev = ttk.Button(self.view.tb_fr_btn, text="Prev")
        self.tb_btn_prev.grid(row=0,rowspan=2,column=col, padx=2)
        col += 1
        self.tb_btn_next = ttk.Button(self.view.tb_fr_btn, text="Next")
        self.tb_btn_next.grid(row=0,rowspan=2,column=col, padx=2)



from ... import logger

def args_mode(parser):
    pass

class Mode:
    def __init__(self, view):
        self.log = logger.new(f"tkGUI.view.{type(self).__name__}")
        self.view = view

    def draw_tb(self, parent, col=0):
        return col
    def draw_cl(self, parent, row=0):
        return row

    def finish_tb(self):
        pass
    def finish_cl(self):
        pass



"""Controller for LIVE sink"""
import argparse
import datetime as dt
import tkinter as tk

import numpy as np
import pysdrlib

from ...utils import args
from .sink import Sink, args_sink

from .dispatch import CMD
# from .arc.plot_base import define_args as freq_args
from .panels import PanelController, PanelChild, Panel
from .plot_base import FreqPlotController, BlitPlot

from .swept import plots

from ...utils.time import strfmt_td

class SinkConfig:
    pass

def args_live(parser):
    sink = args.get_group(parser, "Sink (LIVE)")
    args_sink(sink)

class SinkLive(Sink):
    def __init__(self, ctrl):
        super().__init__(ctrl)

        self.ctrl.view.sink.cl_cb_dev.config(values=list([d for d in pysdrlib.devices.ls()]))
        self.ctrl.view.cb_dev.bind("<<ComboboxSelected>>", self.handle_event)

        self.ctrl.view.sink.cl_ent_rx_rf.bind("<Return>", self.handle_event)
        self.ctrl.view.sink.cl_ent_rx_if.bind("<Return>", self.handle_event)
        self.ctrl.view.sink.cl_ent_rx_bb.bind("<Return>", self.handle_event)
        self.ctrl.view.sink.cl_var_rx_rf.set(str(self.ctrl.model.sink.get_rx_rf()))
        self.ctrl.view.sink.cl_var_rx_if.set(str(self.ctrl.model.sink.get_rx_if()))
        self.ctrl.view.sink.cl_var_rx_bb.set(str(self.ctrl.model.sink.get_rx_bb()))
        self.toggle_gains()

    def start(self):
        if self.ctrl.model.sink.dev is None:
            print("sink is not fully initialized, failed to get device")
            return
        self.ctrl.view.tb_btn_start.config(state=tk.DISABLED)
        self.ctrl.view.tb_btn_stop.config(state=tk.ACTIVE)
        self.ctrl.dispatch.send(CMD.START)

    def stop(self):
        self.ctrl.view.tb_btn_stop.config(state=tk.DISABLED)
        self.ctrl.view.tb_btn_start.config(state=tk.ACTIVE)
        self.ctrl.dispatch.send(CMD.STOP)

    def reset(self):
        self.ctrl.dispatch.send(CMD.RESET)
        self.stop()

    def draw_cl(self):
        self.ctrl.view.sink.cl_var_dev.set(str(type(self.ctrl.model.sink.dev).__name__))
        self.ctrl.view.sink.cl_var_rx_rf.set(str(self.ctrl.model.sink.get_rx_rf()))
        self.ctrl.view.sink.cl_var_rx_if.set(str(self.ctrl.model.sink.get_rx_if()))
        self.ctrl.view.sink.cl_var_rx_bb.set(str(self.ctrl.model.sink.get_rx_bb()))

    def toggle_gains(self):
        if not self.ctrl.model.sink.has_rx_rf():
            self.ctrl.view.sink.cl_fr_rx_rf.grid_forget()
        else:
            self.ctrl.view.sink.cl_fr_rx_rf.grid()
        if not self.ctrl.model.sink.has_rx_if():
            self.ctrl.view.sink.cl_fr_rx_if.grid_forget()
        else:
            self.ctrl.view.sink.cl_fr_rx_if.grid()
        if not self.ctrl.model.sink.has_rx_bb():
            self.ctrl.view.sink.cl_fr_rx_bb.grid_forget()
        else:
            self.ctrl.view.sink.cl_fr_rx_bb.grid()

    # --- GUI bind events and setters --- #
    def handle_event(self, event):
        if event.widget == self.ctrl.view.sink.cl_cb_dev:
            self.set_device(self.ctrl.view.sink.cl_var_dev.get())
        elif event.widget == self.ctrl.view.sink.cl_ent_rx_rf:
            self.set_rx_rf(self.ctrl.view.sink.cl_var_rx_rf.get())
        elif event.widget == self.ctrl.view.sink.cl_ent_rx_if:
            self.set_rx_if(self.ctrl.view.sink.cl_var_rx_if.get())
        elif event.widget == self.ctrl.view.sink.cl_ent_rx_bb:
            self.set_rx_bb(self.ctrl.view.sink.cl_var_rx_bb.get())

    def set_device(self, device):
        self.ctrl.model.sink.set_device(device)
        print(f"Device set to '{self.ctrl.model.sink.name}'")
        self.draw_tb()
        self.draw_cl()
        self.toggle_gains()

    def set_rx_rf(self, gain):
        try:
            gain = int(gain)
            self.ctrl.model.sink.set_rx_rf(gain)
        except ValueError:
            pass
        self.draw_cl()

    def set_rx_if(self, gain):
        try:
            gain = int(gain)
            self.ctrl.model.sink.set_rx_if(gain)
        except ValueError:
            pass
        self.draw_cl()

    def set_rx_bb(self, gain):
        try:
            gain = int(gain)
            self.ctrl.model.sink.set_rx_bb(gain)
        except ValueError:
            pass
        self.draw_cl()



import tkinter as tk
from tkinter import ttk

class PanelView:
    def __init__(self, master):
        self.master = master
        self.main = ttk.PanedWindow(master, orient=tk.VERTICAL)
        self.main.pack(fill=tk.BOTH, expand=True)

        self.btn_row = ttk.Button(master, text="+ ROW", style="AddRow.TButton")
        self.btn_row.place(relx=1,rely=0, x=-100, anchor=tk.NE, bordermode=tk.OUTSIDE, height=30, width=60)

    def update_layout(self):
        self.main.update_idletasks()
        self.master.update_idletasks()

class PanelChild:
    def __init__(self, parent, master):
        self.parent = parent
        self.master = master
        self.root = ttk.Frame(master)
        self.root.pack(fill=tk.BOTH, expand=True)
        self.main = ttk.PanedWindow(self.root, orient=tk.HORIZONTAL)
        self.main.pack(fill=tk.BOTH, expand=True)

        self.btn_col = ttk.Button(master, text="+ COL", style="AddCol.TButton")
        self.btn_col.place(relx=1,rely=0, x=-40, y=0, anchor=tk.NE, bordermode=tk.OUTSIDE, height=30, width=60)
        self.btn_close = ttk.Button(master, text="X", style="Close.TButton")
        self.btn_close.place(relx=1, rely=0, anchor=tk.NE, bordermode=tk.OUTSIDE, height=30, width=30)

    def update_layout(self):
        self.main.update_idletasks()
        self.master.update_idletasks()

class Panel:
    def __init__(self, parent, master):
        self.parent = parent
        self.master = master
        self.root = ttk.Frame(master)
        self.root.pack(fill=tk.BOTH, expand=True)

        self.wgts = {}
        self.sets = {}

        self.fr_sets = ttk.Frame(self.root, borderwidth=1, relief=tk.RAISED)
        self.fr_sets.pack(side=tk.LEFT, fill=tk.Y)

        self.lbl_sets = ttk.Label(self.fr_sets, text="Settings")
        self.lbl_sets.pack(side=tk.TOP, fill=tk.X)

        self.var_view = tk.StringVar(self.fr_sets)
        self.cb_view = ttk.Combobox(self.fr_sets, textvariable=self.var_view, width=20)
        self.cb_view.pack()
        ttk.Separator(self.fr_sets, orient=tk.HORIZONTAL).pack(fill=tk.X, pady=5)

        self.settings = ttk.Frame(self.fr_sets)
        self.settings.pack(side=tk.TOP, fill=tk.BOTH)
        self.fr_sets.pack_forget()

        self.fr_main = ttk.Frame(self.root)
        self.fr_main.pack(fill=tk.BOTH, expand=True)

        self.btn_toggle = ttk.Button(master, text="Settings", style="Settings.TButton")
        self.btn_toggle.place(relx=0,rely=1, x=-5, y=5, anchor=tk.SW)
        self.btn_close = ttk.Button(master, text="X", style="Close.TButton")
        self.btn_close.place(relx=1, rely=0, y=5, anchor=tk.NE, bordermode=tk.OUTSIDE, height=30, width=30)



"""tkGUI Controller"""
import math

from .dispatch import Dispatch, CMD

from ...utils import args
from ..base import Controller as _Controller

from ... import logger
from ...model.model import Model
from ...obj import Frequency
from ...view.tkGUI.base import View as GUI

from ...backend.mpl.plot import BlitPlot
from ...backend.tk import theme as theme_tk

from .mode import Mode
from .sink import Sink
from .panels import PanelController
from .plot_base import TimePlotController, FreqPlotController, BlitPlot

class ControllerConfig:
    theme = "Dark"
    sweep = 50.0
    show = 50.0

def args_ctrl(parser):
    ctrl = args.get_group(parser, "View (tkGUI)")
    ctrl.add_argument("--theme", default=ControllerConfig.theme, choices=[k for k in theme_tk.theme.keys()])
    ctrl.add_argument("--sweep", default=ControllerConfig.sweep, type=float)
    ctrl.add_argument("--show", default=ControllerConfig.show, type=float)
    return ctrl

class Controller(_Controller):
    """tkGUI Controller"""
    def __init__(self, model: Model, view: GUI, mode, sink, **kwargs):
        self.log = logger.new("tkGUI")
        super().__init__(model, view)
        self.view: GUI = self.view # type hints
        self.dispatch: Dispatch = Dispatch(self)
        self.mode: Mode = mode(self)
        self.sink: Sink = sink(self)

        self.view.tb_btn_start.config(command=self.sink.start)
        self.view.tb_btn_stop.config(command=self.sink.stop)
        self.view.tb_btn_reset.config(command=self.sink.reset)

        self.time_show = kwargs.get("show", ControllerConfig.show)
        self.model.set_sweep_time(kwargs.get("sweep", ControllerConfig.sweep))

        self.nfft_exp = int(math.log2(self.model.get_nfft()))
        self._last_f = None
        self.panel: PanelController = None # type: ignore

        self.view.tb_var_sweep.set(f"{self.model.get_sweep_time():02.3f}")
        self.view.tb_ent_sweep.bind("<Return>", self.handle_event)
        self.view.tb_var_show.set(f"{self.time_show:02.3f}")
        self.view.tb_ent_show.bind("<Return>", self.handle_event)

        self.view.tb_var_draw_time.set(f"{0.0:06.3f}s")
        self.view.tb_btn_start.config(command=self.sink.start)
        self.view.tb_btn_stop.config(command=self.sink.stop)
        self.view.tb_btn_reset.config(command=self.sink.reset)

        self.view.cl_ent_fs.bind("<Return>", self.handle_event)
        self.view.cl_ent_cf.bind("<Return>", self.handle_event)
        self.view.cl_ent_nfft_exp.bind("<Return>", self.handle_event)

        self.dispatch.start()

        style = kwargs.get("theme", ControllerConfig.theme)
        theme_tk.get(style)(self.view.root) # pyright: ignore[reportCallIssue]

        self.set_cf(None)
        self.set_fs(None)
        self.set_nfft(None)

        self.view.root.protocol("WM_DELETE_WINDOW", self.quit)
        self.draw()

    def quit(self):
        self.dispatch.stop()
        self.view.root.quit()
        self.view.root.destroy()

    def draw(self):
        self.draw_tb()
        self.draw_cl()
        self.draw_view()

    def draw_tb(self):
        # self.view.var_sweep.set(f"{self.model.get_sweep_time():02.3f}")
        # self.view.var_show.set(f"{self.time_show:02.3f}")
        self.sink.draw_tb()
        self.mode.draw_tb()

    def draw_cl(self):
        # self.view.var_fs.set(str(self.model.get_fs()))
        # self.view.var_cf.set(str(self.model.get_cf()))
        # self.view.lbl_nfft.configure(text=str(self.model.get_nfft()))
        # self.view.var_nfft_exp.set(str(self.nfft_exp))
        self.view.cl_lbl_block_size.configure(text=str(self.model.get_block_size()))
        self.view.cl_lbl_sweep_samples.configure(text=str(self.model.get_sweep_samples()))
        self.sink.draw_cl()
        self.mode.draw_cl()

    def draw_view(self):
        pass

    # --- GUI bind events and setters --- #
    def handle_event(self, event):
        if event.widget == self.view.tb_ent_sweep:
            self.set_time_sweep(self.view.tb_var_sweep.get())
        elif event.widget == self.view.tb_ent_show:
            self.set_time_show(self.view.tb_var_show.get())
        elif event.widget == self.view.cl_ent_fs:
            self.set_fs(self.view.cl_var_fs.get())
        elif event.widget == self.view.cl_ent_cf:
            self.set_cf(self.view.cl_var_cf.get())
        elif event.widget == self.view.cl_ent_nfft_exp:
            self.set_nfft(self.view.cl_var_nfft_exp.get())

    def set_time_sweep(self, ts):
        self.log.trace("set_time_sweep(%s)", ts)
        try:
            self.model.set_sweep_time(float(ts))
        except ValueError:
            pass
        self.view.tb_var_sweep.set(f"{self.model.get_sweep_time():02.3f}")
        self.draw_cl()
    def set_time_show(self, ts):
        self.log.trace("set_time_show(%s)", ts)
        try:
            ts = float(ts)
            self.time_show = ts
        except ValueError:
            pass
        self.view.tb_var_show.set(f"{self.time_show:02.3f}")

    def set_fs(self, fs):
        self.log.trace("set_fs(%s)", fs)
        if not fs is None:
            try:
                Frequency.get(fs)
                self.model.set_fs(fs)
                self.dispatch.send(CMD.UPDATE_FS)
                self.dispatch.send(CMD.UPDATE_F)
            except ValueError:
                pass
        self.view.cl_var_fs.set(str(self.model.sink.get_fs()))
        self.draw_tb()
        self.draw_cl()
    def set_cf(self, cf):
        self.log.trace("set_cf(%s)", cf)
        if not cf is None:
            try:
                Frequency.get(cf)
                self.model.set_cf(cf)
                self.dispatch.send(CMD.UPDATE_F)
            except ValueError:
                pass
        self.view.cl_var_cf.set(str(self.model.sink.get_cf()))
        self.draw_cl()
    def set_nfft(self, exp):
        self.log.trace("set_nfft(%s)", exp)
        if not exp is None:
            try:
                exp = int(exp)
                self.nfft_exp = exp
                self.model.set_nfft(2**exp)
            except ValueError:
                pass
        self.view.cl_var_nfft_exp.set(str(self.nfft_exp))
        self.view.cl_lbl_nfft.config(text=str(self.model.get_nfft()))
        self.dispatch.send(CMD.UPDATE_NFFT)
        self.dispatch.send(CMD.UPDATE_F)
        self.draw_cl()



import tkinter as tk
import tkinter.ttk as ttk

class Scale(ttk.Frame):
    def __init__(self, master=None, **kw):
        ttk.Frame.__init__(self, master)
        self.columnconfigure(0, weight=1)

        from_ = kw.pop("from_", 0)
        to = kw.pop("to", 10)

        self._var = kw.pop("variable", tk.IntVar(master))
        self._var.set(from_)
        self._last_valid = from_

        self.label = ttk.Label(self)
        self.scale = ttk.Scale(self, variable=self._var, from_=from_, to=to, **kw)
        self.scale.bind('<<RangeChanged>>', self._adjust)

        self.scale.pack(side=tk.BOTTOM, fill=tk.X)

        dummy = ttk.Label(self)
        dummy.pack(side=tk.TOP)
        dummy.lower()
        self.label.place(anchor=tk.N)

        self.__tracecb = self._var.trace_add('write', self._adjust)
        self.bind('<Configure>', self._adjust)
        self.bind('<Map>', self._adjust)

    def _adjust(self, *args):
        def adjust_label():
            self.update_idletasks()

            x, y = self.scale.coords()
            y = self.scale.winfo_y() - self.label.winfo_reqheight()
            self.label.place_configure(x=x, y=y)
        from_ = ttk._to_number(self.scale["from"]) # type: ignore
        to = ttk._to_number(self.scale["to"]) # type: ignore
        if to < from_:
            from_, to = to, from_
        newval = self._var.get()
        if not from_ <= newval <= to:
            self.value = self._last_valid
            return
        self._last_valid = newval
        self.label["text"] = newval
        self.after_idle(adjust_label)



"""Initialize CUI Controller"""
from ..base import Controller as _Controller

from ...view.cui import View as CUI
from ...model.model import Model

from .plot_base import PSD

def define_args(parser):
    ctrl = parser.add_argument_group("CUI")
    return ctrl

class Controller(_Controller):
    """CUI Controller"""
    def __init__(self, model: Model, view: CUI, **kwargs):
        super().__init__(model, view)
        self.view: CUI = self.view # type hints
        self.view.menu = self.menu
        self.psd = PSD(self)

    def show_help(self, ind=2):
        print(" "*ind + "h|help  .  .  .  .  .  .  .  : view this message")
        print(" "*ind + "q|quit|exit   .  .  .  .  .  : exit")
        print(" "*ind + "c|config   .  .  .  .  .  .  : show/edit configuration")
        self.cmd_config([], ind=ind+2)
        print(" "*ind + "s|state .  .  .  .  .  .  .  : change state")
        self.cmd_state([], ind=ind+2)
        print(" "*ind + "v|view")
        self.cmd_view([], ind=ind+2)


    def menu(self):
        args = input("pyspecan > ").split(" ")
        if args[0] in ("h", "help"):
            self.show_help()
        elif args[0] in ("q", "quit", "exit"):
            self.view.running = False
        elif args[0] in ("c", "config", "configure"):
            self.cmd_config(args[1:])
        elif args[0] in ("v", "view"):
            self.cmd_view(args[1:])

    def cmd_config(self, args, ind=4):
        def show_help():
            print(" "*ind + "s|show   .  .  .  .  .  .  : view args")
            print(" "*ind + "f|path <path>  .  .  .  .  : set sample rate")
            print(" "*ind + "d|fmt <fmt> .  .  .  .  .  : set sample rate")
            print(" "*ind + "")
            print(" "*ind + "fs <Fs>  .  .  .  .  .  .  : set sample rate")
            print(" "*ind + "cf <cf>  .  .  .  .  .  .  : set center frequency")
            print(" "*ind + "n|nfft <nfft>  .  .  .  .  : set sample rate")
            print(" "*ind + "st|sweep_time <sweep>   .  : set sample rate")

        if len(args) == 0 or args[0] in ("h", "help"):
            show_help()
            return
        # if args[0] in ("s", "show"):
        #     print(" "*ind + f"  {self.model.cur_time():.2f}s/{self.model.tot_time():.2f}s")
        #     print(" "*ind + "  Reader:")
        #     self.model.reader.show(ind+2)

        #     print(f"  Fs: {self.model.Fs} | cf: {self.model.cf}")
        #     print(f"  nfft: {self.model.nfft} | Sweep time: {self.model.sweep_time}")
        # elif args[0] in ("f", "path"):
        #     if len(args) == 2:
        #         self.model.reader.path = args[1]
        #         print(" "*ind + str(self.model.reader.path))
        # elif args[0] in ("d", "fmt") and len(args) == 2:
        #     self.model.reader.fmt = args[1]
        #     print(" "*ind + str(self.model.reader.fmt))
        # elif args[0] in ("fs",) and len(args) == 2:
        #     self.model.Fs = args[1]
        #     print(" "*ind + str(self.model.Fs))
        # elif args[0] in ("cf",) and len(args) == 2:
        #     self.model.cf = args[1]
        #     print(" "*ind + str(self.model.cf))
        # elif args[0] in ("n", "nfft") and len(args) == 2:
        #     self.model.nfft = args[1]
        #     print(" "*ind + str(self.model.nfft))
        # elif args[0] in ("st", "sweep_time") and len(args) == 2:
        #     self.model.sweep_time = args[1]
        #     print(" "*ind + str(self.model.sweep_time))

    def cmd_state(self, args, ind=4):
        def show_help():
            print(" "*ind + "n|next   .  .  .  .  .  .  : advance to next block")
            print(" "*ind + "p|prev   .  .  .  .  .  .  : reverse to previous block")

        if len(args) == 0 or args[0] in ("h", "help"):
            show_help()
            return
        # elif args[0] in ("n", "next"):
        #     valid = self.model.next()
        #     print(" "*ind + f"{valid}")
        # elif args[0] in ("p", "prev"):
        #     valid = self.model.prev()
        #     print(" "*ind + f"{valid}")

    def cmd_view(self, args, ind=4):
        def show_help():
            print(" "*ind + "p|psd .  .  .  .  .  .  .  : view psd")

        if len(args) == 0 or args[0] in ("h", "help"):
            show_help()
            return
        elif args[0] in ("p", "psd"):
            self.psd.plot(self.model.f, self.model.psd())
            self.psd.show()



import importlib

def GetController(view, mode, sink):
    return importlib.import_module(f".controller.{view.path}", "pyspecan").GetController(mode, sink)

def ControllerArgs(view, mode, sink, parser):
    return importlib.import_module(f".controller.{view.path}", "pyspecan").ControllerArgs(mode, sink, parser)



import argparse
import tkinter as tk
from tkinter import ttk

import numpy as np

import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

from ..panels import Panel
from ..plot_base import TimePlotController

from ....utils import ComplexPMF
from ....backend.mpl.plot import _Plot, Plot, BlitPlot

class PlotConfig:
    pass

def define_args(parser: argparse.ArgumentParser):
    pass
class IQ(TimePlotController):
    def __init__(self, parent, pane: Panel, **kwargs):
        pane.master.config(text="IQ")
        self.y_max = 0
        self.x_arr = np.arange(0, dtype=np.float16)
        super().__init__(parent, pane, **kwargs)
        fig = plt.figure(figsize=(5,5), layout="constrained")

        fig.canvas = FigureCanvasTkAgg(fig, master=self.fr_canv)
        self.plotter = BlitPlot(fig) # type: ignore
        self.plotter.canvas.draw()
        self.plotter.canvas.get_tk_widget().pack(side=tk.LEFT, fill=tk.BOTH, expand=True) # type: ignore
        self.plotter.add_ax("iq", fig.add_subplot())

        self.plotter.ax("iq").ax.set_autoscale_on(False)
        self.plotter.ax("iq").ax.grid(True, alpha=0.2)

        self.update()

    def reset(self):
        self.y_max = 0

    def _plot(self, samps):
        prev_max = self.y_max
        self.y_max = np.max((self.y_max, -np.min(samps.real), np.max(samps.real), -np.min(samps.imag), np.max(samps.imag)))
        if not self.y_max == prev_max:
            self.plotter.ax("iq").set_ylim(-self.y_max, self.y_max)
        s_len = len(samps)
        if not s_len == len(self.x_arr):
            self.x_arr = np.arange(s_len) / s_len
            self.plotter.ax("iq").set_xlim(0, 1)

        line_real = self.plotter.ax("iq").plot(self.x_arr, samps.real, name="real")
        line_imag = self.plotter.ax("iq").plot(self.x_arr, samps.imag, name="imag")

        self.update()

    def draw_settings(self, row=0):
        pass



from enum import Enum

class Sink(Enum):
    """Sink modes"""
    NONE = "none"
    FILE = "file"
    LIVE = "live"

    @classmethod
    def choices(cls):
        """return mode choices"""
        arr = []
        for inst in cls:
            if inst == cls.NONE:
                continue
            arr.append(inst.name)
        return arr

    @classmethod
    def get(cls, option: str):
        for inst in cls:
            if inst.name == option:
                return inst
        return cls.NONE

class Mode(Enum):
    """Specan modes"""
    NONE = "none"
    SWEPT = "swept"
    RT = "rt"

    @classmethod
    def choices(cls):
        """return mode choices"""
        arr = []
        for inst in cls:
            if inst == cls.NONE:
                continue
            arr.append(inst.name)
        return arr

    @classmethod
    def get(cls, option: str):
        for inst in cls:
            if inst.name == option:
                return inst
        return cls.NONE

class View(Enum):
    """Specan views"""
    NONE = (("none",), None)
    CUI = (("c", "cui"), "cui")
    tkGUI = (("tk", "tkgui"), "tk_gui")

    @property
    def path(self):
        """return view import path name"""
        return self.value[1]

    @classmethod
    def get(cls, option: str):
        """return view instance"""
        for inst in cls:
            if inst.name == option or option in inst.value[0]:
                return inst
        return cls.NONE

    @classmethod
    def choices(cls):
        """return view choices"""
        arr = []
        for inst in cls:
            if inst == cls.NONE:
                continue
            for i in inst.value[0]:
                arr.append(i)
            arr.append(inst.name)
        return arr


class config:
    """Global variables for convenience"""
    __instance = None

    SENTINEL = object()
    SINK: Sink = Sink.NONE
    MODE: Mode = Mode.NONE

    MON_MEM: bool = False
    PROFILE: bool = False
    PROFILE_PATH = [None, "pyspecan.prof"][1]

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
            cls.__instance._init()
        return cls.__instance

    def _init(self):
        pass



"""tkGUI SWEPT mode"""
import argparse

from ...utils import args
from .mode import Mode, args_mode

from .panels import PanelController, PanelChild, Panel
from .plot_base import FreqPlotController, BlitPlot

from .swept import plots

class ModeConfig:
    psd = True
    spg = False

def args_swept(parser: argparse.ArgumentParser):
    mode = args.get_group(parser, "Mode (SWEPT)")
    args_mode(mode)
    # freq_args(parser)
    # mode = parser.add_argument_group("SWEPT mode")
    # mode.add_argument("--psd", action="store_false", help="show psd")
    # mode.add_argument("--spg", action="store_true", help="show spectrogram")

class ModeSwept(Mode):
    def __init__(self, ctrl, **kwargs):
        super().__init__(ctrl)
        self.panel = PanelController(self.ctrl, self.ctrl.view.panel, plots)
        child = self.panel.rows[0]
        pane = self.panel.cols[child][0]
        pane.var_view.set("PSD")
        self.panel.set_view(None, child, pane)



"""File reader"""
import pathlib
from enum import Enum

import numpy as np
from numpy._typing._array_like import NDArray
import pysdrlib
from pysdrlib import Formats as _Formats
from pysdrlib import File as _File

from ...utils import args
from .sink import Sink, args_sink

from ... import err

class Format(Enum):
    ci8  = _Formats.ci8
    ci16 = _Formats.ci16
    cf32 = _Formats.cf32
    cf64 = _Formats.cf64
    cf128 = _Formats.cf128

    @classmethod
    def choices(cls):
        return [inst.name for inst in cls]

def args_file(parser):
    sink = args.get_group(parser, "Sink (FILE)")
    args_sink(sink)
    sink.add_argument("-f", "--path", default=None, help="file path")
    sink.add_argument("-d", "--fmt", choices=Format.choices(), default=Format.cf32.name, help="data format")

class SinkFile(Sink):
    __slots__ = (
        "dev", "fmt"
    )
    def __init__(self, model, **kwargs):
        path = kwargs.get("path", None)
        fmt = kwargs.get("fmt", Format.cf32.name)
        super().__init__(model, **kwargs)
        self.dev = _File()

        self.set_fmt(fmt)
        self.set_path(path)
        self.dev.open()

    def show(self, ind=0):
        print(" "*ind + f"{self.percent():06.2f}% [{self.dev.fmt.name}] {self.dev.path}")
        print(" "*ind + f"{self.dev.cur_samp}/{self.dev.max_samp}")

    def get_path(self):
        return self.dev.get_path()
    def set_path(self, path):
        if path is None or path == "":
            return
        path = pathlib.Path(path)

        if not path.exists():
            return
        self.dev.set_path(path)
        self.reset()

    def get_fmt(self):
        return self.fmt # TODO: is this needed?
    def set_fmt(self, fmt):
        self.fmt = Format[fmt]
        self.dev.set_fmt(fmt)

    def _set_fs(self, fs):
        return fs
    def _set_cf(self, cf):
        return cf

    def reset(self):
        self.dev.reset()

    def next(self, count: int):
        # self.log.trace("next(%s)", count)
        try:
            samples = self.dev.next(count)
        except pysdrlib.file.err.Overflow:
            return False
        self._samples = samples
        # self._psd = None
        return True

    def prev(self, count: int):
        try:
            samples = self.dev.prev(count)
        except pysdrlib.file.err.Overflow:
            return False
        self._samples = samples
        # self._psd = None
        return True

    def forward(self, count: int):
        return self.dev.forward(count)

    def reverse(self, count: int):
        return self.dev.reverse(count)

    def percent(self):
        return self.dev.percent()

    def cur_time(self):
        return self.dev.cur_samp/self.get_fs()
    def tot_time(self):
        return self.dev.max_samp/self.get_fs()
    def skip_time(self, s):
        samps = int(self.model.Fs * s)
        # print(f"Skipping {s:.3f}s, {samps} ({samps/self.reader.max_samp*100:.2f}%)")
        self.dev.cur_samp += samps

    @property
    def cur_samp(self):
        return self.dev.cur_samp

    @property
    def max_samp(self):
        return self.dev.max_samp



"""tkGUI base PlotController"""
import argparse
import time

import tkinter as tk
from tkinter import ttk
import numpy as np

from .panels import Panel

from ...utils.window import WindowLUT
from ...utils.psd import psd
from ...utils.stft import psd as psd_rt

from ...backend.mpl.plot import _Plot, Plot, BlitPlot

def args_plot(parser, cfg):
    parser.add_argument("-rl", "--ref_level", default=cfg.ref_level, type=float, help="ref Level")
    parser.add_argument("-sd", "--scale_div", default=cfg.scale_div, type=float, help="scale per division")
    parser.add_argument("-vb", "--vbw", default=cfg.vbw, type=float, help="video bandwidth")
    parser.add_argument("-w", "--window", default=cfg.window, choices=[k for k in WindowLUT.keys()], help="FFT window function")

class _PlotController:
    """Controller for view.tkGUI.GUIPlot"""
    __slots__ = (
        "parent", "pane", "plotter",
        "ready", "enabled",
        "btn_enable", "lbl_time", "fr_canv"
    )
    def __init__(self, parent, pane: Panel, **kwargs):
        self.parent = parent
        self.pane = pane
        self.ready = False
        self.enabled = True
        self.btn_enable = ttk.Button(pane.master, text="DISABLE", style="Toggle.TButton", command=self.toggle)
        self.btn_enable.place(relx=1, rely=0, x=-100, y=5, anchor=tk.N, bordermode=tk.OUTSIDE, height=30, width=80)
        self.lbl_time = ttk.Label(pane.master, text="00.000s", style="Time.TLabel")
        self.lbl_time.place(relx=1, rely=0, x=-180, y=5, anchor=tk.N, bordermode=tk.OUTSIDE, height=20, width=60)
        self.plotter: Plot = None # type: ignore

        self.fr_canv = ttk.Frame(pane.fr_main)
        self.fr_canv.pack(fill=tk.BOTH, expand=True)

    def toggle(self, *args, **kwargs):
        self.enabled = not self.enabled
        if self.enabled:
            self.btn_enable.config(text="DISABLE")
        else:
            self.btn_enable.config(text="ENABLE")

    def update(self):
        """Update view plot"""
        self.plotter.update()

    def update_fs(self, fs):
        """Update sample rate"""

    def reset(self):
        """Reset plot view"""
        pass

    def plot(self, samps, **kwargs):
        """Update plot data"""
        if not self.enabled:
            return None
        ptime = time.perf_counter()
        self._plot(samps, **kwargs)
        ptime = time.perf_counter() - ptime
        try: # catch threading state mismatch
            self.lbl_time.config(text=f"{ptime:06.3f}s")
        except tk.TclError:
            pass
        return ptime

    def _plot(self, samps, **kwargs):
        raise NotImplementedError()

    def draw_settings(self, row=0):
        """Initialize settings panel"""
        raise NotImplementedError()

class TimePlotController(_PlotController):
    """Controller for view.tkGUI time-domain plots"""
    def __init__(self, parent, pane: Panel, **kwargs):
        super().__init__(parent, pane)
        self.draw_settings()

    def update(self):
        self.plotter.canvas.draw()

    def _plot(self, samps): # type: ignore
        raise NotImplementedError()

    def draw_settings(self, row=0):
        pass

class FreqPlotController(_PlotController):
    """Controller for view.tkGUI frequency-domain plots"""
    __slots__ = (
        "window", "vbw", "scale", "ref_level",
        "lbl_lo", "lbl_hi"
    )
    def __init__(self, parent, pane: Panel, **kwargs):
        super().__init__(parent, pane)

        self.window =  kwargs.get("window", "blackman")
        self.vbw = kwargs.get("vbw", 10.0)
        self.scale = kwargs.get("scale_div", 10.0)
        self.ref_level = kwargs.get("ref_level", 0.0)
        self.draw_settings()

        self.set_ref_level(self.pane.sets["ref_level"].get())

        self.lbl_lo = ttk.Label(self.pane.fr_main, text="V")
        self.lbl_hi = ttk.Label(self.pane.fr_main, text="^")
        self.lbl_lo.lift()
        self.lbl_hi.lift()

    def update(self):
        self.plotter.canvas.draw()

    def update_f(self, f):
        """Set plot xticks and xlabels"""

    def update_nfft(self, nfft):
        """Update plot nfft"""

    def _plot(self, samps): # type: ignore
        raise NotImplementedError()

    def psd(self, samps):
        vbw = self.vbw
        if vbw <= 0:
            vbw = None
        return psd(samps, self.parent.model.sink.get_fs().raw, vbw, self.window)

    @property
    def y_top(self):
        """Return plot maximum amplitude"""
        return self.ref_level
    @property
    def y_btm(self):
        """Return plot minimum amplitude"""
        return self.ref_level - (10*self.scale)

    def _show_y_location(self, psd):
        if np.all(psd < self.y_btm):
            self.lbl_lo.place(relx=0.2, rely=0.9, width=20, height=20)
        else:
            if self.lbl_lo.winfo_ismapped():
                self.lbl_lo.place_forget()
        if np.all(psd > self.y_top):
            self.lbl_hi.place(relx=0.2, rely=0.1, width=20, height=20)
        else:
            if self.lbl_hi.winfo_ismapped():
                self.lbl_hi.place_forget()

    def draw_settings(self, row=0):
        var_scale = tk.StringVar(self.pane.settings, str(self.scale))
        ent_scale = ttk.Entry(self.pane.settings, textvariable=var_scale, width=10)
        ent_scale.bind("<Return>", self.handle_event)

        var_ref_level = tk.StringVar(self.pane.settings, str(self.ref_level))
        ent_ref_level = ttk.Entry(self.pane.settings, textvariable=var_ref_level, width=10)
        ent_ref_level.bind("<Return>", self.handle_event)

        var_vbw = tk.StringVar(self.pane.settings, str(self.vbw))
        ent_vbw = ttk.Entry(self.pane.settings, textvariable=var_vbw, width=10)
        ent_vbw.bind("<Return>", self.handle_event)

        var_window = tk.StringVar(self.pane.settings, str(self.window))
        cb_window = ttk.Combobox(self.pane.settings, textvariable=var_window, width=9)
        cb_window.configure(values=[k for k in WindowLUT.keys()])
        cb_window.bind("<<ComboboxSelected>>", self.handle_event)

        self.pane.wgts["scale"] = ent_scale
        self.pane.sets["scale"] = var_scale
        self.pane.wgts["ref_level"] = ent_ref_level
        self.pane.sets["ref_level"] = var_ref_level
        self.pane.wgts["vbw"] = ent_vbw
        self.pane.sets["vbw"] = var_vbw
        self.pane.wgts["window"] = cb_window
        self.pane.sets["window"] = var_window

        ttk.Label(self.pane.settings, text="Scale/Div").grid(row=row, column=0)
        ent_scale.grid(row=row, column=1)
        row += 1
        ttk.Label(self.pane.settings, text="Ref Level").grid(row=row, column=0)
        ent_ref_level.grid(row=row, column=1)
        row += 1
        ttk.Label(self.pane.settings, text="VBW").grid(row=row, column=0)
        ent_vbw.grid(row=row, column=1)
        row += 1
        ttk.Label(self.pane.settings, text="Window").grid(row=row, column=0)
        cb_window.grid(row=row, column=1)
        row += 1
        return row

    # --- GUI bind events and setters --- #
    def handle_event(self, event):
        if event.widget == self.pane.wgts["scale"]:
            self.set_scale(self.pane.sets["scale"].get())
        elif event.widget == self.pane.wgts["ref_level"]:
            self.set_ref_level(self.pane.sets["ref_level"].get())
        elif event.widget == self.pane.wgts["vbw"]:
            self.set_vbw(self.pane.sets["vbw"].get())
        elif event.widget == self.pane.wgts["window"]:
            self.set_window(self.pane.sets["window"].get())

    def set_scale(self, scale):
        """set plot scale"""
        try:
            scale = float(scale)
            self.scale = scale
        except ValueError:
            scale = self.scale
        self.pane.sets["scale"].set(str(self.scale))

    def set_ref_level(self, ref):
        """Set plot ref level"""
        try:
            ref = float(ref)
            self.ref_level = ref
        except ValueError:
            ref = self.ref_level
        self.pane.sets["ref_level"].set(str(self.ref_level))

    def set_vbw(self, smooth):
        """Set plot vbw"""
        try:
            smooth = float(smooth)
            if smooth <= 0.0:
                smooth = 0.0
            self.vbw = smooth
        except ValueError:
            smooth = self.vbw
        self.pane.sets["vbw"].set(str(self.vbw))
    def set_window(self, window):
        """Set plot window function"""
        self.window = window

    def _calc_ref_level(self, psd, mul=0.1):
        pmax = np.max(psd)
        pmax = int(pmax * (1-mul)) if pmax < 0 else int(pmax * (1+mul))
        return pmax

class FreqPlotControllerRT(FreqPlotController):
    def __init__(self, parent, pane: Panel, **kwargs):
        self.overlap = kwargs.get("overlap", )
        super().__init__(parent, pane, **kwargs)

    def psd(self, samps):
        vbw = self.vbw
        if vbw <= 0:
            vbw = None
        return psd_rt(samps, self.parent.model.get_nfft(), self.parent.model.mode.get_overlap(), self.parent.model.get_fs().raw, vbw, self.window)

    def _plot(self, samps):
        raise NotImplementedError()

    # --- GUI bind events and setters --- #



"""Create a tkGUI view"""

from .. import err

from ..config import Mode, Sink
from .tkGUI.base import View, args_view
from .tkGUI.mode_rt import ModeRT, args_rt
from .tkGUI.mode_swept import ModeSwept, args_swept
from .tkGUI.sink_file import SinkFile, args_file
from .tkGUI.sink_live import SinkLive, args_live

def GetMode(mode):
    if mode == Mode.SWEPT:
        return ModeSwept
    elif mode == Mode.RT:
        return ModeRT
    raise err.UnknownOption(f"Unknown mode: {mode}")

def GetSink(sink):
    if sink == Sink.FILE:
        return SinkFile
    elif sink == Sink.LIVE:
        return SinkLive
    raise err.UnknownOption(f"Unknown sink: {sink}")

def GetView(mode, sink):
    m: Mode = GetMode(mode) # type: ignore
    s: Sink = GetSink(sink) # type: ignore
    return lambda m=m, s=s, **kwargs: View(m, s, **kwargs)

def ViewArgs(mode, sink, parser):
    args_view(parser)
    if mode == Mode.SWEPT:
        args_swept(parser)
    elif mode == Mode.RT:
        args_rt(parser)

    if sink == Sink.FILE:
        args_file(parser)
    elif sink == Sink.LIVE:
        args_live(parser)



# freq plots
from .psd import PSD
from .spg import SPG
from .spg3d import SPG3D
from .freq import Freq

# time plots
from ..shared.iq import IQ
from ..shared.pmf import PMF
from ..shared.iq3d import IQ3D
from ..shared.phasor import Phasor

plots = {
    "PSD": PSD,
    "Spectrogram": SPG,
    # "SPG 3D": SPG3D,
    "FFT": Freq,

    "IQ": IQ,
    "IQ 3D": IQ3D,
    "PMF": PMF,
    "Phasor": Phasor
}



class SI:
    __slots__ = ("_val", "_mul", "_si")
    _units = {
        0: ("",""),
        3: ("",""),
        6: ("",""),
        9: ("","")
    }
    def __init__(self, val: float, mul=None):
        if mul is None:
            self._val: float = val
            self._mul: int = 0
            self._si: float = 0
            self._calc()
        else:
            self._mul: int = mul
            self._val: float = val * self._factor()
            self._si: float = val

    def __str__(self):
        return self.to_str()

    def __repr__(self):
        return f"{type(self).__name__} ({self._si}, v{self._val}, m{self._mul})"

    @classmethod
    def get(cls, val):
        if isinstance(val, str):
            return cls.fm_str(val)
        elif isinstance(val, cls):
            return val
        else:
            return cls(val)

    def to_str(self):
        fmt = f"{self._si:.3f}"
        unit = self._units[self._mul][1]
        if not unit == "":
            fmt += f"{unit}"
        return fmt

    @classmethod
    def fm_str(cls, val: str):
        for mul, fmt in cls._units.items():
            if mul == 0:
                continue
            if fmt[0] in val:
                v = val.split(fmt[0])[0]
                return cls(float(v), mul)
            elif fmt[1] in val:
                v = val.split(fmt[1])[0]
                return cls(float(v), mul)
        return cls(float(val))

    @property
    def value(self):
        return self._si

    @property
    def raw(self):
        return self._val

    @raw.setter
    def raw(self, val):
        self._val = val
        self._calc()

    def _factor(self, mul=None):
        if mul is None:
            mul = self._mul
        return 10**mul

    def _calc(self):
        mul = 0
        v = abs(self._val)
        if v > 0:
            while v > 1000:
                v = v/1000
                mul += 3
        self._mul = mul
        self._si = self._val / self._factor(self._mul)

    def __add__(self, other):
        if issubclass(other, SI):
            return self.raw + other.raw # type: ignore
        else:
            return self.raw + other
    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, SI):
            return self.raw - other.raw # type: ignore
        else:
            return self.raw - other
    def __rsub__(self, other):
        if isinstance(other, SI):
            return other.raw - self.raw
        else:
            return other - self.raw
    def __mul__(self, other):
        if isinstance(other, SI):
            return self.raw * other.raw # type: ignore
        else:
            return self.raw * other
    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, SI):
            return self.raw / other.raw # type: ignore
        else:
            return self.raw / other
    def __rtruediv__(self, other):
        if isinstance(other, SI):
            return other.raw / self.raw # type: ignore
        else:
            return other / self.raw


class Frequency(SI):
    _units = {
        0: ("Hz",""),
        3: ("kHz","k"),
        6: ("MHz","M"),
        9: ("GHz","G")
    }

if __name__ == "__main__":
    unit = Frequency(1_000_000)
    print(f"original: {unit}")
    freq_str = str(unit)
    new_unit = Frequency.fm_str(freq_str)
    print(f"from_str: {new_unit}")
    print(f"from freq: {Frequency.get(unit)}")



"""
pyspecan

A spectrum analyzer library

Github: https://github.com/Anonoei/pyspecan

PyPI: https://pypi.org/project/pyspecan/

This is intended to be used as a callable module (or a script), but can be used as a library
"""

__version__ = "0.5.2"
__author__ = "Anonoei <to+dev@an0.cx>"

from . import _internal

from .config import config, Mode, View
from . import err
from . import obj
from . import utils

# from .model.base import Model
from .model.base import Model
from .model.model import GetModel

from .specan import SpecAn



"""tkGUI RT mode"""
import argparse

from ...utils import args
from .mode import Mode, args_mode
from .plot_base import args_plot

from .panels import PanelController, PanelChild, Panel
from .plot_base import FreqPlotController, BlitPlot

from .rt import plots
from ...backend.mpl.color import cmap

class ModeConfig:
    ref_level = 0
    scale_div = 10.0
    vbw = 0.0
    window = "blackman"
    x = 1001
    y = 600
    cmap = "hot"

def args_rt(parser: argparse.ArgumentParser):
    mode = args.get_group(parser, "Mode (RT)")
    args_mode(mode)
    mode.add_argument("--x", default=ModeConfig.x, type=int, help="histogram x pixels")
    mode.add_argument("--y", default=ModeConfig.y, type=int, help="histogram y pixels")
    mode.add_argument("--cmap", default=ModeConfig.cmap, choices=[k for k in cmap.keys()], help="histogram color map")

class ModeRT(Mode):
    def __init__(self, ctrl, **kwargs):
        super().__init__(ctrl)

        self.ctrl.view.mode.cl_var_overlap.set(f"{self.ctrl.model.mode.get_overlap():.2f}")
        self.ctrl.view.mode.cl_ent_overlap.bind("<Return>", self.handle_event)
        self.panel = PanelController(self.ctrl, self.ctrl.view.panel, plots)
        child = self.panel.rows[0]
        pane = self.panel.cols[child][0]
        pane.var_view.set("Persistent Histogram")
        self.panel.set_view(None, child, pane)

    # --- GUI bind events and setters --- #
    def handle_event(self, event):
        if event.widget == self.ctrl.view.mode.cl_ent_overlap:
            self.set_overlap(self.ctrl.view.mode.cl_var_overlap.get())

    def set_overlap(self, overlap):
        try:
            overlap = float(overlap)
            self.ctrl.model.mode.set_overlap(overlap)
        except ValueError:
            pass
        self.ctrl.view.mode.cl_var_overlap.set(f"{self.ctrl.model.mode.get_overlap():.2f}")



import argparse

import numpy as np

from ...utils import args
from .mode import Mode, args_mode

from ...utils import psd as _psd

def args_swept(parser: argparse.ArgumentParser):
    mode = args.get_group(parser, "Mode (SWEPT)")
    args_mode(mode)

class ModeSwept(Mode):
    def __init__(self, model, **kwargs):
        super().__init__(model)



import argparse
import tkinter as tk
from tkinter import ttk

import numpy as np

import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

from ..panels import Panel
from ..plot_base import TimePlotController

from ....utils import ComplexPMF
from ....backend.mpl.plot import _Plot, Plot, BlitPlot

class PlotConfig:
    pass

def define_args(parser: argparse.ArgumentParser):
    pass
class PMF(TimePlotController):
    def __init__(self, parent, pane: Panel, **kwargs):
        pane.master.config(text="PMF")
        self.pmf = ComplexPMF(256)
        super().__init__(parent, pane, **kwargs)
        fig = plt.figure(figsize=(5,5), layout="constrained")

        fig.canvas = FigureCanvasTkAgg(fig, master=self.fr_canv)
        self.plotter = BlitPlot(fig) # type: ignore
        self.plotter.canvas.draw()
        self.plotter.canvas.get_tk_widget().pack(side=tk.LEFT, fill=tk.BOTH, expand=True) # type: ignore
        self.plotter.add_ax("pmf", fig.add_subplot())

        self.plotter.ax("pmf").ax.grid(True, alpha=0.2)

        self.update()

    def reset(self):
        self.pmf = ComplexPMF(256)

    def _plot(self, samps):
        self.pmf.update(samps)

        x = self.pmf.x
        y = self.pmf.y

        line_real = self.plotter.ax("pmf").plot(y.real, x.real, name="real")
        line_imag = self.plotter.ax("pmf").plot(y.imag, x.imag, name="imag")

        self.update()

    def draw_settings(self, row=0):
        pass



import tkinter as tk
import tkinter.ttk as ttk

from ...backend.tk import widgets

from ...utils import args
from .mode import Mode, args_mode

def args_rt(parser):
    mode = args.get_group(parser, "Mode (RT)")
    args_mode(mode)

class ModeRT(Mode):
    __slots__ = (
        # control panel
        "cl_var_overlap", "cl_ent_overlap"
    )
    def __init__(self, view, **kwargs):
        super().__init__(view, **kwargs)
        self.cl_var_overlap: tk.StringVar = None # type: ignore
        self.cl_ent_overlap: ttk.Entry = None # type: ignore

    def draw_tb(self, parent, col=0):
        return col
    def draw_cl(self, parent, row=0):
        self.cl_var_overlap = tk.StringVar(parent)
        ttk.Label(parent, text="Overlap:").grid(row=row,column=0)
        self.cl_ent_overlap = ttk.Entry(parent, textvariable=self.cl_var_overlap, width=4)
        self.cl_ent_overlap.grid(row=row,column=1, sticky=tk.NSEW)
        row += 1
        return row



"""tk themes"""
from ._theme import _Theme

class Light(_Theme):
    """pyspecan tk light theme"""
    C_FG = "#000000"
    C_FG_TEXT = "#000000"
    C_FG_ALT = C_FG
    C_FG_ACTIVE = C_FG
    C_FG_DISABLED = "#888888"
    C_FG_PRESSED = "#880000"
    C_FG_SELECTED = C_FG
    C_FG_FOCUS = C_FG
    C_FG_READONLY = "#444444"
    C_FG_INVALID = "#AA11AA"

    C_BG = "#EEEEEE"
    C_BG_TEXT = "#FFFFFF"
    C_BG_ALT = "#FFFFFF"
    C_BG_ACTIVE = "#DDDDDD"
    C_BG_DISABLED = "#CCCCCC"
    C_BG_PRESSED = "#AAAAAA"
    C_BG_SELECTED = C_BG
    C_BG_FOCUS = C_BG
    C_BG_READONLY = "#CCCCCC"
    C_BG_INVALID = C_BG

class Dark(_Theme):
    """pyspecan tk dark theme"""
    C_FG = "#FFFFFF"
    C_FG_TEXT = "#FFFFFF"
    C_FG_ALT = C_FG
    C_FG_ACTIVE = C_FG
    C_FG_DISABLED = "#888888"
    C_FG_PRESSED = "#FF8888"
    C_FG_SELECTED = C_FG
    C_FG_FOCUS = C_FG
    C_FG_READONLY = "#CCCCCC"
    C_FG_INVALID = "#AA11AA"

    C_BG = "#111111"
    C_BG_TEXT = "#222222"
    C_BG_ALT = "#000000"
    C_BG_ACTIVE = "#222222"
    C_BG_DISABLED = "#000000"
    C_BG_PRESSED = "#888888"
    C_BG_SELECTED = C_BG
    C_BG_FOCUS = C_BG
    C_BG_READONLY = "#333333"
    C_BG_INVALID = C_BG

theme = {
    "Light": Light,
    "Dark": Dark
}

def get(name: str) -> _Theme:
    return theme[name]



import argparse

import numpy as np

from .. import logger
from ...obj import Frequency

class SinkConfig:
    Fs = 1
    cf = 0

def args_sink(parser):
    parser.add_argument("-fs", "--Fs", default=SinkConfig.Fs, type="frequency", help="sample rate")
    parser.add_argument("-cf", "--cf", default=SinkConfig.cf, type="frequency", help="center frequency")

class Sink:
    __slots__ = (
        "log", "model",
        "_samples", "_cf", "_Fs"
    )
    def __init__(self, model, **kwargs):
        self.log = logger.new(f"model.{type(self).__name__}")
        self.model = model
        self._cf: Frequency = Frequency.get(kwargs.get("cf", SinkConfig.cf))
        self._Fs: Frequency = Frequency.get(kwargs.get("Fs", SinkConfig.Fs))
        self._samples = np.empty((), dtype=np.complex64)

    def reset(self):
        self._samples = np.empty(self.model.mode.get_block_size(), dtype=np.complex64)

    @property
    def samples(self):
        return self._samples

    def show(self, ind=0):
        raise NotImplementedError()

    def set_fs(self, fs):
        self.log.debug("set_fs(%s)", fs)
        if isinstance(fs, str):
            fs = Frequency.get(fs).raw
        fs = self._set_fs(fs)
        self._Fs = Frequency.get(fs)
        self.model.f = self.model.update_f()
    def get_fs(self):
        return self._Fs
    def set_cf(self, cf):
        self.log.debug("set_cf(%s)", cf)
        if isinstance(cf, str):
            cf = Frequency.get(cf).raw
        cf = self._set_cf(cf)
        self._cf = Frequency.get(cf)
        self.model.f = self.model.update_f()

    def get_cf(self):
        return self._cf

    # --- Defined in child classes --- #
    def _set_fs(self, fs):
        raise NotImplementedError()
    def _set_cf(self, cf):
        raise NotImplementedError()



import matplotlib.pyplot as plt

from ...backend.mpl.plot import Plot

class CUIPlot:
    def __init__(self, parent, fig, plotter=Plot):
        self.parent = parent
        self.plotter = plotter(fig)

    @property
    def fig(self):
        return self.plotter.fig
    def ax(self, name):
        return self.plotter.ax(name)
    def add_ax(self, *args, **kwargs):
        return self.plotter.add_ax(*args,**kwargs)

class PSD(CUIPlot):
    def __init__(self, parent):
        fig = plt.figure(figsize=(5,5), layout="constrained")
        super().__init__(parent, fig)
        self.plotter.add_ax("psd", fig.add_subplot())

    def plot(self, freq, psd):
        self.ax("psd").ax.set_title("PSD")
        self.ax("psd").plot(freq, psd, name="psd", color="y")

    def show(self):
        self.plotter.fig.show()



import argparse

from ..obj import Frequency

def update_register(parser: argparse.ArgumentParser):
    parser.register("type", "frequency", Frequency.get)

def get_group(
        parser: argparse.ArgumentParser,
        title: str, description=None, **kwargs):
    """Runs parser.add_argument_group, getting existing group if it has the same name"""
    if parser._action_groups:
        for group in parser._action_groups:
            if group.title == title:
                return group
    return parser.add_argument_group(title, description, **kwargs)



"""tkinter dialogs"""
import pathlib

from tkinter import filedialog

def get_file(strict=True, title: str="", filetypes=None):
    """Prompts for file selection"""
    while True:
        if filetypes is None:
            file = filedialog.askopenfilename(title=title)
        else:
            file = filedialog.askopenfilename(title=title, filetypes=filetypes)
        if file == "" or file == ():
            if strict:
                continue
            return None
        file = pathlib.Path(file)
        if not file.exists():
            if strict:
                continue
            return None
        return file



import queue
import threading
import time
from enum import Enum, auto

from ... import logger
from ...config import config, Mode
from ...utils.monitor import Memory

class CMD(Enum):
    NEXT = auto()
    PREV = auto()

    START = auto()
    STOP = auto()
    RESET = auto()

    PLOT = auto()

    UPDATE_F = auto()
    UPDATE_NFFT = auto()
    UPDATE_FS = auto()

    UPDATE_MODE = auto()
    UPDATE_SINK = auto()

class STATE(Enum):
    WAITING = auto()
    RUNNING = auto()

class Dispatch:
    def __init__(self, controller):
        self.log = logger.new("tkGUI.dispatch")
        self.ctrl = controller
        self.queue = queue.Queue()

        self.state = STATE.WAITING
        self.running = True
        self.thread = threading.Thread(target=self._run, name="dispatcher")

        self._last_f = self._get_f()

    def start(self):
        self.running = True
        self.thread.start()

    def stop(self):
        self.send(CMD.STOP)
        self.running = False
        self.thread.join(timeout=1)

    def send(self, cmd: CMD, val=None):
        self.queue.put((cmd, val))

    def _run(self):
        while self.running:
            if self.state is not STATE.RUNNING:
                if self.queue.qsize() == 0:
                    time.sleep(0.2)
                    continue
            else:
                self._loop()
            if not self.queue.qsize() == 0:
                cmd, val = self.queue.get()
                if cmd is CMD.NEXT:
                    self.log.trace("executing CMD.NEXT")
                    if self.state is STATE.RUNNING:
                        self.send(CMD.STOP)
                    self._next()
                elif cmd is CMD.PREV:
                    self.log.trace("executing CMD.PREV")
                    if self.state is STATE.RUNNING:
                        self.send(CMD.STOP)
                    self._prev()
                elif cmd is CMD.START:
                    self.log.trace("executing CMD.START")
                    self.state = STATE.RUNNING
                elif cmd is CMD.STOP:
                    self.log.trace("executing CMD.STOP")
                    self.state = STATE.WAITING
                elif cmd is CMD.PLOT:
                    self.log.trace("executing CMD.PLOT")
                    pass
                elif cmd is CMD.RESET:
                    self.log.trace("executing CMD.RESET")
                    if self.state is STATE.RUNNING:
                        self.state = STATE.WAITING
                    self.ctrl.mode.panel.on_reset()
                    self.ctrl.model.reset()
                    self.ctrl.draw_tb()
                elif cmd is CMD.UPDATE_F:
                    self.log.trace("executing CMD.UPDATE_F")
                    self._last_f = None
                    self._update_f()
                elif cmd is CMD.UPDATE_NFFT:
                    self.log.trace("executing CMD.UPDATE_NFFT")
                    self.ctrl.mode.panel.on_update_nfft(self.ctrl.model.get_nfft())
                elif cmd is CMD.UPDATE_FS:
                    self.log.trace("executing CMD.UPDATE_FS")
                    self.ctrl.mode.panel.on_update_fs(self.ctrl.model.get_fs())

    def on_plot(self):
        ptime = time.perf_counter()
        # self._update_f()
        self.ctrl.mode.panel.on_plot(self.ctrl.model)

        ptime = (time.perf_counter() - ptime)
        self.ctrl.view.tb_var_draw_time.set(f"{ptime:06.3f}s")
        self.ctrl.draw_tb()

        # print(f"Plotted in {ptime*1000:.1f}ms / {self.time_show}")
        return ptime

    def _loop(self):
        time_show = self.ctrl.time_show/1000 # convert ms to s
        valid, ptime = self._next()
        if not valid or ptime is None:
            self.send(CMD.STOP)
            return
        wait = time_show-ptime
        if wait > 0:
            self.ctrl.view.tb_lbl_msg.configure(text="")
            time.sleep(wait)
        else:
            if not self.ctrl.model.mode.get_sweep_time() == 0.0:
                if config.MODE == Mode.SWEPT:
                    self.ctrl.model.skip_time(-wait)
                self.ctrl.view.tb_lbl_msg.configure(text="OVERFLOW")

    def _prev(self): # fails on Sink.LIVE
        valid = self.ctrl.model.sink.prev(self.ctrl.model.mode.get_block_size())
        tplot = None
        if valid:
            tplot = self.on_plot()
        return (valid, tplot)

    def _next(self):
        valid = self.ctrl.model.sink.next(self.ctrl.model.mode.get_block_size())
        tplot = None
        if valid:
            tplot = self.on_plot()
        return (valid, tplot)

    def _get_f(self):
        return (self.ctrl.model.f[0], self.ctrl.model.f[-1]+(self.ctrl.model.f[-1]-self.ctrl.model.f[-2]), len(self.ctrl.model.f))
    def _update_f(self):
        if self._last_f is None:
            self._last_f = self._get_f()
            self.ctrl.mode.panel.on_update_f(self._last_f)
        elif not self.ctrl.model.f[0] == self._last_f[0] and not len(self.ctrl.model.f) == self._last_f[2]:
            self._last_f = self._get_f()
            self.ctrl.mode.panel.on_update_f(self._last_f)

    @property
    def last_f(self):
        return self._last_f



import tkinter as tk
import tkinter.ttk as ttk

from ...backend.tk import widgets

from ...utils import args
from .sink import Sink, args_sink

def args_live(parser):
    sink = args.get_group(parser, "Sink (LIVE)")
    args_sink(sink)

class SinkLive(Sink):
    __slots__ = (
        # control panel
        "cl_var_dev", "cl_cb_dev",
        "cl_fr_rx_rf", "cl_var_rx_rf", "cl_ent_rx_rf",
        "cl_fr_rx_if", "cl_var_rx_if", "cl_ent_rx_if",
        "cl_fr_rx_bb", "cl_var_rx_bb", "cl_ent_rx_bb"
    )
    def __init__(self, view, **kwargs):
        super().__init__(view, **kwargs)
        self.cl_var_dev: tk.StringVar = None # type: ignore
        self.cl_cb_dev: ttk.Combobox = None # type: ignore
        self.cl_fr_rx_rf: ttk.Frame = None # type: ignore
        self.cl_var_rx_rf: tk.StringVar = None # type: ignore
        self.cl_ent_rx_rf: ttk.Entry = None # type: ignore
        self.cl_fr_rx_if: ttk.Frame = None # type: ignore
        self.cl_var_rx_if: tk.StringVar = None # type: ignore
        self.cl_ent_rx_if: ttk.Entry = None # type: ignore
        self.cl_fr_rx_bb: ttk.Frame = None # type: ignore
        self.cl_var_rx_bb: tk.StringVar = None # type: ignore
        self.cl_ent_rx_bb: ttk.Entry = None # type: ignore

    def draw_cl(self, parent, row=0):
        parent.columnconfigure(1, weight=1)
        self.cl_var_dev = tk.StringVar(parent) # TODO: switch from ent to combo
        ttk.Label(parent, text="Device:").grid(row=row,column=0, sticky=tk.W)
        self.cl_cb_dev = ttk.Combobox(parent, textvariable=self.cl_var_dev, width=20)
        # self.ent_dev = ttk.Entry(root, textvariable=self.var_dev, state=tk.DISABLED, width=10)
        self.cl_cb_dev.grid(row=row,column=1, sticky=tk.NSEW)
        row += 1
        ttk.Separator(parent, orient=tk.HORIZONTAL).grid(row=row, column=0, columnspan=2, pady=5, sticky=tk.EW)
        row += 1
        self.cl_fr_rx_rf = ttk.Frame(parent)
        self.cl_fr_rx_rf.columnconfigure(1, weight=1)
        self.cl_var_rx_rf = tk.StringVar(self.cl_fr_rx_rf)
        ttk.Label(self.cl_fr_rx_rf, text="Rx RF Gain:").grid(row=0,column=0, sticky=tk.W)
        self.cl_ent_rx_rf = ttk.Entry(self.cl_fr_rx_rf, textvariable=self.cl_var_rx_rf, width=3)
        self.cl_ent_rx_rf.grid(row=0,column=1, sticky=tk.E)
        self.cl_fr_rx_rf.grid(row=row,column=0,columnspan=2, sticky=tk.NSEW)
        # root.columnconfigure(row, weight=1)
        row += 1
        self.cl_fr_rx_if = ttk.Frame(parent)
        self.cl_fr_rx_if.columnconfigure(1, weight=1)
        self.cl_var_rx_if = tk.StringVar(self.cl_fr_rx_if)
        ttk.Label(self.cl_fr_rx_if, text="Rx IF Gain:").grid(row=0,column=0, sticky=tk.W)
        self.cl_ent_rx_if = ttk.Entry(self.cl_fr_rx_if, textvariable=self.cl_var_rx_if, width=3)
        self.cl_ent_rx_if.grid(row=0,column=1, sticky=tk.E)
        self.cl_fr_rx_if.grid(row=row,column=0,columnspan=2, sticky=tk.NSEW)
        # root.columnconfigure(row, weight=1)
        row += 1
        self.cl_fr_rx_bb = ttk.Frame(parent)
        self.cl_fr_rx_bb.columnconfigure(1, weight=1)
        self.cl_var_rx_bb = tk.StringVar(self.cl_fr_rx_bb)
        ttk.Label(self.cl_fr_rx_bb, text="Rx BB Gain:").grid(row=0,column=0, sticky=tk.W)
        self.cl_ent_rx_bb = ttk.Entry(self.cl_fr_rx_bb, textvariable=self.cl_var_rx_bb, width=3)
        self.cl_ent_rx_bb.grid(row=0,column=1, sticky=tk.E)
        self.cl_fr_rx_bb.grid(row=row,column=0,columnspan=2, sticky=tk.NSEW)
        # root.columnconfigure(row, weight=1)
        return row



from .pst import PST
from .spg import SPG

# time plots
from ..shared.iq import IQ
from ..shared.pmf import PMF
from ..shared.iq3d import IQ3D
from ..shared.phasor import Phasor

plots = {
    "Persistent Histogram": PST,
    "Spectrogram": SPG,

    "IQ": IQ,
    "IQ 3D": IQ3D,
    "PMF": PMF,
    "Phasor": Phasor
}



import argparse
import tkinter as tk
from tkinter import ttk

import numpy as np

import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

from ..panels import Panel
from ..plot_base import TimePlotController

from ....utils import ComplexPMF
from ....backend.mpl.plot import _Plot, Plot, BlitPlot

class PlotConfig:
    pass

def define_args(parser: argparse.ArgumentParser):
    pass
class Phasor(TimePlotController):
    def __init__(self, parent, pane: Panel, **kwargs):
        pane.master.config(text="Phasor")
        super().__init__(parent, pane, **kwargs)
        fig = plt.figure(figsize=(5,5), layout="constrained")

        fig.canvas = FigureCanvasTkAgg(fig, master=self.fr_canv)
        self.plotter = BlitPlot(fig) # type: ignore
        self.plotter.canvas.draw()
        self.plotter.canvas.get_tk_widget().pack(side=tk.LEFT, fill=tk.BOTH, expand=True) # type: ignore
        self.plotter.add_ax("iq", fig.add_subplot())

        self.plotter.ax("iq").ax.set_autoscale_on(False)
        self.plotter.ax("iq").ax.grid(True, alpha=0.2)
        self.plotter.ax("iq").set_xlim(-1, 1)
        self.plotter.ax("iq").set_ylim(-1, 1)
        self.plotter.ax("iq").ax.set_box_aspect(1)

        self.update()

    def reset(self):
        self.y_max = 0

    def _plot(self, samps):
        s_max = np.max((samps.real, samps.imag))
        self.pane.master.config(text=f"Phasor (norm {s_max:03.2f})")
        samps = samps / s_max

        self.plotter.ax("iq").scatter(
            samps.real, samps.imag, name="iq",
            linestyle=None, marker=".", linewidth=0,
        )

        self.update()

    def draw_settings(self, row=0):
        pass



import argparse
import tkinter as tk
from tkinter import ttk

import numpy as np

import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

from ..panels import Panel
from ..plot_base import FreqPlotControllerRT

from ....utils import vbw as _vbw
from ....obj import Frequency
from ....utils.window import WindowLUT
from ....backend.mpl.plot import _Plot, Plot, BlitPlot

class PlotConfig:
    ref_level = 0
    scale_div = 10.0
    vbw = 10.0
    window = "blackman"

def define_args(parser: argparse.ArgumentParser):
    parser.add_argument("-rl", "--ref_level", default=PlotConfig.ref_level, type=float, help="ref Level")
    parser.add_argument("-sd", "--scale_div", default=PlotConfig.scale_div, type=float, help="scale per division")
    parser.add_argument("-vb", "--vbw", default=PlotConfig.vbw, type=float, help="video bandwidth")
    parser.add_argument("-w", "--window", default=PlotConfig.window, choices=[k for k in WindowLUT.keys()], help="FFT window function")

class SPG(FreqPlotControllerRT):
    def __init__(self, parent, pane: Panel, **kwargs):
        pane.master.config(text="Spectrogram")
        self.max_count = 256
        self.psds: np.ndarray = None # type: ignore
        self.fmin = None
        self.fmax = None
        if not "ref_level" in kwargs:
            kwargs["ref_level"] = PlotConfig.ref_level
        if not "scale_div" in kwargs:
            kwargs["scale_div"] = PlotConfig.scale_div
        if not "vbw" in kwargs:
            kwargs["vbw"] = PlotConfig.vbw
        if not "window" in kwargs:
            kwargs["vbw"] = PlotConfig.vbw
        super().__init__(parent, pane, **kwargs)
        self.reset()
        fig = plt.figure(figsize=(5,5), layout="constrained")

        fig.canvas = FigureCanvasTkAgg(fig, master=self.fr_canv)
        self.plotter = BlitPlot(fig) # type: ignore
        self.plotter.canvas.draw()
        self.plotter.canvas.get_tk_widget().pack(side=tk.LEFT, fill=tk.BOTH, expand=True) # type: ignore
        self.plotter.add_ax("spg", fig.add_subplot())

        self.plotter.ax("spg").ax.set_autoscale_on(False)
        self.plotter.ax("spg").ax.locator_params(axis="x", nbins=5)
        self.plotter.ax("spg").ax.locator_params(axis="y", nbins=5)

        self.update_f(self.parent.dispatch.last_f)
        self.set_y()
        self.update()

    def reset(self):
        self.psds = np.zeros((self.max_count, self.parent.model.get_nfft()), dtype=np.float32)
        self.psds[:,:] = -np.inf

    def _plot(self, samps):
        psd = self.psd(samps)
        psd = np.sum(psd, axis=1) / psd.shape[1]
        psd = np.clip(psd, self.y_btm, self.y_top)
        y_rng = abs(self.y_btm - self.y_top)
        psd += (0 - self.y_btm)
        psd = psd / y_rng

        self.psds = np.roll(self.psds, 1, axis=0)
        self.psds[0,:] = psd
        # print(self.psds.shape)
        im = self.plotter.ax("spg").imshow(
            self.psds, name="spg",
            aspect="auto", origin="upper", cmap="magma",
            vmin=0.0, vmax=1.0,
            interpolation="nearest", resample=False, rasterized=True
        )
        self.update()

    def update_f(self, f):
        fmin, fmax, fnum = f
        if self.fmin == fmin and self.fmax == fmax:
            return
        else:
            self.fmin = fmin
            self.fmax = fmax
        psd_tick = np.linspace(fmin, fmax, 5)
        psd_text = [str(Frequency.get(f)) for f in psd_tick]

        spg_tick = np.linspace(0, fnum+1, 5)
        spg_text = psd_text
        self.plotter.ax("spg").ax.set_xlim(0, fnum)
        self.plotter.ax("spg").ax.set_xticks(spg_tick, spg_text)

    def update_nfft(self, nfft):
        self.psds = np.zeros((self.max_count, nfft), dtype=np.float32)
        self.psds[:,:] = -np.inf
        self.reset()

    def set_y(self):
        """Set plot ylimits"""
        self.plotter.ax("spg").set_ylim(self.max_count, 0)

    def draw_settings(self, row=0):
        row = super().draw_settings(row)
        var_max_count = tk.StringVar(self.pane.settings, str(self.max_count))
        ent_max_count = ttk.Entry(self.pane.settings, textvariable=var_max_count, width=10)
        ent_max_count.bind("<Return>", self.handle_event)

        self.pane.wgts["max_count"] = ent_max_count
        self.pane.sets["max_count"] = var_max_count

        ttk.Separator(self.pane.settings, orient=tk.HORIZONTAL).grid(row=row,column=0,columnspan=3, pady=5, sticky=tk.EW)
        row += 1
        ttk.Label(self.pane.settings, text="Spectrogram").grid(row=row, column=0,columnspan=2)
        row += 1
        ttk.Label(self.pane.settings, text="Max Count").grid(row=row, column=0)
        ent_max_count.grid(row=row, column=1)
        row += 1
        return row

    # --- GUI bind events and setters --- #
    def handle_event(self, event):
        if event.widget == self.pane.wgts["max_count"]:
            self.set_count(self.pane.sets["max_count"].get())
        else:
            super().handle_event(event)

    def set_count(self, count):
        prev = int(self.max_count)
        try:
            count = int(count)
            self.max_count = count
        except ValueError:
            count = self.max_count
        self.pane.sets["count"].set(str(self.max_count))
        if not prev == self.max_count:
            self.reset()

    def set_scale(self, scale):
        prev = float(self.scale)
        super().set_scale(scale)
        if not prev == self.scale:
            self.set_y()

    def set_ref_level(self, ref):
        prev = float(self.ref_level)
        super().set_ref_level(ref)
        if not prev == self.ref_level:
            self.set_y()

    def set_vbw(self, smooth):
        prev = float(self.vbw)
        super().set_vbw(smooth)
        if not prev == self.vbw:
            self.psd_min = None
            self.psd_max = None



import argparse
from ...utils import args

from .mode import Mode, args_mode
from ...utils import stft

class ModeConfig:
    overlap = 0.6
    block_max = 1024*1024
    min_fft = 3
    max_fft = 512

def args_rt(parser: argparse.ArgumentParser):
    mode = args.get_group(parser, "Mode (RT)")
    args_mode(mode)
    mode.add_argument("--overlap", default=ModeConfig.overlap, type=float)
    mode.add_argument("--block_max", default=ModeConfig.block_max, type=int)
    mode.add_argument("--min_fft", default=ModeConfig.min_fft, type=int)
    mode.add_argument("--max_fft", default=ModeConfig.max_fft, type=int)

class ModeRT(Mode):
    __slots__ = (
        "_overlap", "_block_max",
        "_min_fft", "_max_fft"
    )
    def __init__(self, model, **kwargs):
        super().__init__(model)
        self._overlap = kwargs.get("overlap", ModeConfig.overlap)
        self._block_max = kwargs.get("block_max", ModeConfig.block_max)
        self._min_fft = kwargs.get("min_fft", ModeConfig.min_fft)
        self._max_fft = kwargs.get("max_fft", ModeConfig.max_fft)
        self.update_blocksize()

    def update_blocksize(self):
        _fs = self.model.get_fs()
        _nfft = self.model.get_nfft()
        overlap = 1-self._overlap

        self._block_size = int(_fs * (self._sweep_time/1000))
        num_fft = int(self._block_size/(_nfft*overlap))

        min_fft_b = int(self._min_fft * _nfft * overlap)
        max_fft_b = int(self._max_fft * _nfft * overlap)
        min_fft_ms = (min_fft_b/_fs)*1000
        max_fft_ms = (max_fft_b/_fs)*1000

        if num_fft < self._min_fft:
            self.log.debug("_block_size must be greater than %s (%s), num_fft %s < min_fft %s", min_fft_b, f"{min_fft_ms:.3f}ms", num_fft, self._min_fft)
            self._block_size = min_fft_b
            super().set_sweep_time(min_fft_ms)
        elif num_fft > self._max_fft:
            self.log.debug("_block_size must be less than %s (%s), num_fft %s < min_fft %s", max_fft_b, f"{max_fft_ms:.3f}ms", num_fft, self._max_fft)
            self._block_size = max_fft_b
            super().set_sweep_time(max_fft_ms)

    def get_overlap(self):
        return self._overlap
    def set_overlap(self, overlap):
        if overlap <= 0.0 or overlap > 1.0:
            raise ValueError
        self._overlap = float(overlap)

    def set_sweep_time(self, ts):
        super().set_sweep_time(ts)
        self.update_blocksize()



import tkinter as tk
from tkinter import ttk

from ... import logger
from ...view.tkGUI.panels import PanelView, PanelChild, Panel

from .plot_base import TimePlotController, FreqPlotController, BlitPlot
from .dispatch import CMD

class PanelController:
    def __init__(self, parent, panel: PanelView, plots):
        self.log = logger.new("tkGUI.Panels")
        self.parent = parent
        self.panel = panel
        self.rows = []
        self.cols = {}
        self.view = {}
        self.active = []
        self.plots = plots

        self.panel.btn_row.configure(command=self.add_row)
        self.add_row()

    def add_row(self):
        if len(self.rows) > 5:
            return
        frame = ttk.LabelFrame(self.panel.main, text=f"Row {len(self.rows)}")
        child = PanelChild(self, frame)
        self.rows.append(child)
        self.cols[child] = []
        self.view[child] = {}

        self.panel.main.add(frame, weight=1)

        child.btn_close.configure(command=lambda c=child: self.del_row(c))
        child.btn_col.configure(command=lambda c=child: self.add_col(c))

        self.add_col(child)
        self.panel.update_layout()

    def del_row(self, child: PanelChild):
        idx = self.rows.index(child)
        del self.cols[child]
        self.panel.main.remove(child.master)
        self.rows.pop(idx).root.destroy()

    def add_col(self, child: PanelChild):
        if len(self.cols[child]) > 5:
            return
        frame = ttk.LabelFrame(child.main, text=f"Col {len(self.cols[child])}")
        pane = Panel(self, frame)
        self.cols[child].append(pane)
        self.view[child][pane] = None

        child.main.add(frame, weight=1)
        #ttk.Label(self.panes[-1].main, text=f"Row {len(self.parent.panes)}, Col {len(self.panes)}").pack()

        pane.btn_close.configure(command=lambda c=child, p=pane: self.del_col(c,p))
        pane.btn_toggle.configure(command=lambda c=child, p=pane: self.toggle_settings(c,p))
        self.toggle_settings(child, pane)

        self.set_settings(child, pane)
        child.update_layout()

    def del_col(self, child: PanelChild, pane: Panel):
        idx = self.cols[child].index(pane)
        child.main.remove(pane.master)
        plot = self.view[child][pane]
        if plot is not None:
            self.active.pop(self.active.index(plot))
        del self.view[child][pane]
        self.cols[child].pop(idx).root.destroy()
        if len(self.cols[child]) == 0:
            self.del_row(child)

    def toggle_settings(self, child: PanelChild, pane: Panel):
        """Toggle settings panel visibility"""
        if pane.fr_sets.winfo_ismapped():
            pane.fr_sets.forget()
            # self.btn_toggle.config(text="Show Settings")
        else:
            pane.fr_sets.pack(side=tk.LEFT, fill=tk.Y, before=pane.fr_main)
            # self.btn_toggle.config(text="Hide Settings")

    def set_settings(self, child: PanelChild, pane: Panel):
        pane.cb_view.config(values=list(k for k in self.plots.keys()))
        pane.cb_view.bind("<<ComboboxSelected>>", lambda e,c=child, p=pane: self.set_view(e,c,p))

    def set_view(self, e, child: PanelChild, pane: Panel):
        view = pane.var_view.get()
        if view in self.plots:
            if self.view[child][pane] is not None:
                pane.wgts = {}
                pane.sets = {}
                for ch in pane.fr_main.winfo_children():
                    ch.destroy()
                for ch in pane.settings.winfo_children():
                    ch.destroy()
                self.del_active(child, pane, self.view[child][pane])
            plot = self.plots[view](self.parent, pane)
            self.view[child][pane] = plot
            self.add_active(child, pane, plot)

    def get_pane(self, child: PanelChild, pane: Panel):
        return self.cols[child][self.cols[child].index(pane)]

    def get_view(self, child: PanelChild, pane: Panel):
        return self.view[child][pane]

    def add_active(self, child, pane, plot):
        self.active.append(plot)

    def del_active(self, child, pane, plot):
        plot.enabled = False
        self.active.pop(self.active.index(plot))

    def on_plot(self, model):
        for view in self.active:
            if not isinstance(view.plotter, BlitPlot):
                view.plotter.cla()
                self.log.info("Cleared plot on %s", type(view).__name__)
            # self.log.trace("Calling plot() on %s", type(view).__name__)
            view.plot(model.samples)

    def on_update_f(self, f):
        for view in self.active:
            if isinstance(view, FreqPlotController):
                self.log.trace("Calling update_f() on %s", type(view).__name__)
                view.update_f(f)

    def on_update_nfft(self, nfft):
        for view in self.active:
            if isinstance(view, FreqPlotController):
                self.log.trace("Calling update_nfft() on %s", type(view).__name__)
                view.update_nfft(nfft)

    def on_update_fs(self, Fs):
        for view in self.active:
            self.log.trace("Calling update_Fs() on %s", type(view).__name__)
            view.update_fs(Fs)

    def on_reset(self):
        for view in self.active:
            self.log.trace("Calling reset() on %s", type(view).__name__)
            view.reset()



import typing
import numpy as np

from .. import err
from .. import logger
from ..obj import Frequency

from .mode import Mode
from .sink import Sink

def args_model(parser):
    parser.add_argument("-n", "--nfft", default=1024, help="FFT size")

class Model:
    __slots__ = (
        "log", "mode", "sink",
        "f", "_psd", "_nfft"
    )
    def __init__(self, mode, sink, **kwargs):
        self.log = logger.new("model")
        self._nfft = int(kwargs.get("nfft", 1024))

        self.sink: Sink = sink(self, **kwargs)
        self.mode: Mode = mode(self, **kwargs)
        self.sink.reset()

        self.f = self.update_f()
        self._psd = np.empty(self._nfft, dtype=np.float32)

    def show(self, ind=0):
        print(" "*ind + f"{type(self).__name__} Sink:")
        self.sink.show(ind+2)

    def reset(self):
        self.log.debug("reset")
        self.mode.reset()
        self.sink.reset()

    @property
    def samples(self):
        return self.sink.samples

    def psd(self, vbw=None, win="blackman") -> np.ndarray:
        ...

    def get_fs(self):
        return self.sink.get_fs()
    def set_fs(self, fs):
        self.sink.set_fs(fs)

    def get_cf(self):
        return self.sink.get_cf()
    def set_cf(self, cf):
        self.sink.set_cf(cf)

    def get_nfft(self):
        return self._nfft
    def set_nfft(self, nfft):
        self.log.debug("set_nfft(%s)", nfft)
        self._nfft = int(nfft)
        self.f = self.update_f()
        self._psd = np.empty(self._nfft, dtype=np.float32)

    def get_block_size(self):
        return self.mode.get_block_size()
    def set_block_size(self, size):
        self.mode.set_block_size(size)

    def get_sweep_time(self):
        return self.mode.get_sweep_time()
    def set_sweep_time(self, ts):
        self.mode.set_sweep_time(ts)

    def get_sweep_samples(self):
        return int(self.sink.get_fs() * (self.mode.get_sweep_time()/1000))

    def update_f(self):
        return np.arange(
            -self.sink.get_fs().raw/2,
            self.sink.get_fs().raw/2,
            self.sink.get_fs().raw/self._nfft
        ) + self.sink.get_cf().raw



"""Initialize Controller base class"""
from ..model.model import Model
from ..view.base import View as _View

class Controller:
    """Parent controller class"""
    __slots__ = (
        "model", "view"
    )
    def __init__(self, model: Model, view: _View):
        self.model: Model = model
        self.view = view



from typing import Callable

from .. import err
from ..config import config
from ..config import Mode as _Mode
from ..config import Sink as _Sink

from .base import Model, args_model
from .mode import Mode
from .mode import ModeSwept, args_swept
from .mode import ModeRT, args_rt
from .sink import Sink
from .sink import SinkFile, args_file
from .sink import SinkLive, args_live

def GetMode(mode):
    if mode == _Mode.SWEPT:
        return ModeSwept
    elif mode == _Mode.RT:
        return ModeRT
    raise err.UnknownOption(f"Unknown mode: {mode}")

def GetSink(sink):
    if sink == _Sink.FILE:
        return SinkFile
    elif sink == _Sink.LIVE:
        return SinkLive
    raise err.UnknownOption(f"Unknown sink: {sink}")

def GetModel(mode, sink):
    mode = GetMode(mode)
    sink = GetSink(sink)
    return lambda m=mode, s=sink, **kwargs: Model(m, s, **kwargs)

def ModelArgs(mode, sink, parser):
    args_model(parser)
    if mode == _Mode.SWEPT:
        args_swept(parser)
    elif mode == _Mode.RT:
        args_rt(parser)

    if sink == _Sink.FILE:
        args_file(parser)
    elif sink == _Sink.LIVE:
        args_live(parser)



import numpy as np

from .. import logger

class ModeConfig:
    sweep_time = 50.0

def args_mode(parser):
    parser.add_argument("-st", "--sweep_time", default=ModeConfig.sweep_time, help="[ms] fft sweep time")

class Mode:
    __slots__ = (
        "log", "model",
        "_sweep_time", "_block_size"
    )
    def __init__(self, model, **kwargs):
        self.log = logger.new(f"model.{type(self).__name__}")
        self.model = model
        self._sweep_time: float = kwargs.get("sweep_time", ModeConfig.sweep_time)
        self._block_size: int = self.model._nfft

    def reset(self):
        pass

    def show(self, ind=0):
        raise NotImplementedError()

    def get_block_size(self):
        return self._block_size
    def set_block_size(self, size):
        self.log.debug("set_block_size(%s)", size)
        self._block_size = size
        self.model.sink.reset() # reset model.sink._samples size
    def get_sweep_time(self):
        return self._sweep_time
    def set_sweep_time(self, ts):
        self.log.debug("set_sweep_time(%s)", ts)
        self._sweep_time = ts



"""Create a GUI view"""
import tkinter as tk
import tkinter.ttk as ttk

from ..base import View as _View
from ...config import config, Mode

from ...backend.tk import widgets
from ...backend.mpl import theme as theme_mpl

from .mode import Mode
from .sink import Sink
from .panels import PanelView

LUT = {
    "fr": ttk.Frame,
    "ent": ttk.Entry,
    "btn": ttk.Button,
    "lbl": ttk.Label,
    "cb": ttk.Combobox,
    "sld": widgets.Scale,

    "var": tk.Variable,
    "vs": tk.StringVar,
    "vi": tk.IntVar,
}

def args_view(parser):
    pass

class View(_View):
    """Parent tkGUI view class"""

    __slots__ = (
        "root", "_main",
        "main", "fr_tb", "fr_view", "fr_ctrl", "frame",
        # toolbar
        "tb_fr_sink", "tb_fr_mode", "tb_fr_time", "tb_fr_btn", "tb_fr_msg",
        "tb_btn_start", "tb_btn_stop", "tb_btn_reset",
        "tb_var_sweep", "tb_ent_sweep",
        "tb_var_show", "tb_ent_show",
        "tb_lbl_msg"
        "tb_var_draw_time", "tb_lbl_draw_time",
        # control panel
        "cl_fr_sink", "cl_fr_mode", "cl_fr_view", "cl_fr_lbls",
        "cl_var_fs", "cl_ent_fs",
        "cl_var_cf", "cl_ent_cf",
        "cl_var_nfft_exp", "cl_ent_nfft_exp",
        "cl_lbl_nfft",
        "cl_lbl_block_size", "cl_lbl_sweep_samples",
    )
    def __init__(self, mode, sink, root=tk.Tk(), **kwargs):
        self.root = root
        self.root.title("pyspecan")

        theme_mpl.get(kwargs.get("theme", "Dark"))() # Set matplotlib theme

        self._main = ttk.Frame(self.root)
        self._main.pack(expand=True, fill=tk.BOTH)

        self.fr_tb = ttk.Frame(self._main, height=20)

        self.main = ttk.PanedWindow(self._main, orient=tk.HORIZONTAL)

        self.fr_view = ttk.Frame(self.main)
        self.panel = PanelView(self.fr_view)

        self.fr_cl = ttk.Frame(self.main, width=100)

        # toolbar
        self.tb_fr_sink = ttk.Frame(self.fr_tb)
        self.tb_fr_mode = ttk.Frame(self.fr_tb)
        self.tb_fr_time = ttk.Frame(self.fr_tb)
        self.tb_var_sweep: tk.StringVar = None # type: ignore
        self.tb_ent_sweep: ttk.Entry = None # type: ignore
        self.tb_var_show: tk.StringVar = None # type: ignore
        self.tb_ent_show: ttk.Entry = None # type: ignore
        self.tb_fr_btn = ttk.Frame(self.fr_tb)
        self.tb_btn_start: ttk.Button = None # type: ignore
        self.tb_btn_stop: ttk.Button = None # type: ignore
        self.tb_btn_reset: ttk.Button = None # type: ignore
        self.tb_fr_msg = ttk.Frame(self.fr_tb)
        self.tb_lbl_msg: ttk.Label = None # type: ignore
        self.tb_var_draw_time: tk.StringVar = None # type: ignore
        self.tb_lbl_draw_time: ttk.Label = None # type: ignore
        # control panel
        self.cl_fr_sink = ttk.Frame(self.fr_cl)
        self.cl_fr_mode = ttk.Frame(self.fr_cl)
        self.cl_fr_view = ttk.Frame(self.fr_cl)
        self.cl_var_fs: tk.StringVar = None # type: ignore
        self.cl_ent_fs: ttk.Entry = None # type: ignore
        self.cl_var_cf: tk.StringVar = None # type: ignore
        self.cl_ent_cf: ttk.Entry = None # type: ignore
        self.cl_var_nfft_exp: tk.StringVar = None # type: ignore
        self.cl_ent_nfft_exp: ttk.Entry = None # type: ignore
        self.cl_lbl_nfft: ttk.Label = None # type: ignore
        self.cl_fr_lbls = ttk.Frame(self.fr_cl)
        self.cl_lbl_block_size: ttk.Label = None # type: ignore
        self.cl_lbl_sweep_samples: ttk.Label = None # type: ignore

        self.mode: Mode = mode(self)
        self.sink: Sink = sink(self)

        self.draw_tb()
        self.sink.finish_tb()
        self.mode.finish_tb()
        self.fr_tb.pack(side=tk.TOP, fill=tk.X)

        ttk.Separator(self._main, orient=tk.HORIZONTAL).pack(fill=tk.X)

        self.main.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        self.draw_cl()
        self.sink.finish_cl()
        self.sink.finish_cl()
        self.main.add(self.fr_cl)

        self.main.add(self.fr_view)

    def draw_tb(self):
        """Draw toolbar frame"""
        self.sink.draw_tb(self.tb_fr_sink)
        self.tb_fr_sink.pack(fill=tk.Y, side=tk.LEFT)
        self.mode.draw_tb(self.tb_fr_mode)
        self.tb_fr_mode.pack(fill=tk.Y, side=tk.LEFT)

        ttk.Separator(self.fr_tb, orient=tk.VERTICAL).pack(fill=tk.Y, side=tk.LEFT)

        self.draw_tb_time(self.tb_fr_time)
        self.tb_fr_time.pack(fill=tk.Y, side=tk.LEFT)

        ttk.Separator(self.fr_tb, orient=tk.VERTICAL).pack(fill=tk.Y, side=tk.LEFT)

        self.draw_tb_btn(self.tb_fr_btn)
        self.tb_fr_btn.pack(fill=tk.Y, side=tk.LEFT, pady=2)

        ttk.Separator(self.fr_tb, orient=tk.VERTICAL).pack(fill=tk.Y, side=tk.LEFT)

        self.draw_tb_msg(self.tb_fr_msg)
        self.tb_fr_msg.pack(fill=tk.BOTH, side=tk.RIGHT)

    def draw_tb_time(self, parent, col=0):
        ttk.Label(parent, text="Sweep").grid(row=0,column=col)
        self.tb_var_sweep = tk.StringVar(parent)
        self.tb_ent_sweep = ttk.Entry(parent, textvariable=self.tb_var_sweep, width=8)
        self.tb_ent_sweep.grid(row=1,column=col, padx=2, pady=2)
        col += 1
        ttk.Label(parent, text="Show").grid(row=0,column=col)
        self.tb_var_show = tk.StringVar(parent)
        self.tb_ent_show = ttk.Entry(parent, textvariable=self.tb_var_show, width=8)
        self.tb_ent_show.grid(row=1,column=col, padx=2, pady=2)
        col += 1
        return col

    def draw_tb_btn(self, parent, col=0):
        parent.grid_rowconfigure(0, weight=1)
        self.tb_btn_start = ttk.Button(parent, text="Start")
        self.tb_btn_start.grid(row=0,rowspan=2,column=col, padx=2, sticky=tk.NS)
        col += 1
        self.tb_btn_stop = ttk.Button(parent, text="Stop", state=tk.DISABLED)
        self.tb_btn_stop.grid(row=0,rowspan=2,column=col, padx=2, sticky=tk.NS)
        col += 1
        self.tb_btn_reset = ttk.Button(parent, text="Reset")
        self.tb_btn_reset.grid(row=0,rowspan=2,column=col, padx=2, sticky=tk.NS)
        col += 1
        return col

    def draw_tb_msg(self, parent, col=0):
        self.tb_lbl_msg = ttk.Label(parent, text="")
        self.tb_lbl_msg.grid(row=0,column=col, rowspan=2, sticky=tk.E)
        parent.grid_columnconfigure(col, weight=1)
        col += 1
        self.tb_var_draw_time = tk.StringVar(parent)
        self.tb_lbl_draw_time = ttk.Label(parent, textvariable=self.tb_var_draw_time)
        ttk.Label(parent, text="Draw").grid(row=0,column=col, sticky=tk.E)
        self.tb_lbl_draw_time.grid(row=1,column=col, sticky=tk.E)
        parent.grid_columnconfigure(col, weight=1)
        return col

    def draw_cl(self):
        """Draw control frame"""
        self.sink.draw_cl(self.cl_fr_sink)
        self.cl_fr_sink.pack(padx=2,pady=2, fill=tk.X)
        self.mode.draw_cl(self.cl_fr_mode)
        self.cl_fr_mode.pack(padx=2,pady=2, fill=tk.X)

        ttk.Separator(self.fr_cl, orient=tk.HORIZONTAL).pack(fill=tk.X, pady=5)

        self.draw_cl_view(self.cl_fr_view)
        self.cl_fr_view.pack(padx=2,pady=2, fill=tk.X)

        ttk.Separator(self.fr_cl, orient=tk.HORIZONTAL).pack(fill=tk.X, pady=5)

        self.draw_cl_lbls(self.cl_fr_lbls)
        self.cl_fr_lbls.pack(padx=2,pady=2, fill=tk.X)

    def draw_cl_view(self, parent, row=0):
        parent.columnconfigure(2, weight=1)
        self.cl_var_fs = tk.StringVar(parent)
        ttk.Label(parent, text="Sample rate:").grid(row=row,column=0, sticky=tk.W)
        self.cl_ent_fs = ttk.Entry(parent, textvariable=self.cl_var_fs, width=10)
        self.cl_ent_fs.grid(row=row,column=1, columnspan=2,sticky=tk.E)
        row += 1
        self.cl_var_cf = tk.StringVar(parent)
        ttk.Label(parent, text="Center freq:").grid(row=row,column=0, sticky=tk.W)
        self.cl_ent_cf = ttk.Entry(parent, textvariable=self.cl_var_cf, width=10)
        self.cl_ent_cf.grid(row=row,column=1, columnspan=2,sticky=tk.E)
        row += 1
        self.cl_var_nfft_exp = tk.StringVar(parent)
        ttk.Label(parent, text="NFFT 2^").grid(row=row,column=0, sticky=tk.W)
        self.cl_ent_nfft_exp = ttk.Entry(parent, textvariable=self.cl_var_nfft_exp, width=2)
        self.cl_ent_nfft_exp.grid(row=row,column=1, sticky=tk.W)
        self.cl_lbl_nfft = ttk.Label(parent)
        self.cl_lbl_nfft.grid(row=row,column=2, sticky=tk.E)
        return row

    def draw_cl_lbls(self, parent, row=0):
        ttk.Label(parent, text="Block size: ").grid(row=row, column=0, sticky=tk.W)
        self.cl_lbl_block_size = ttk.Label(parent)
        self.cl_lbl_block_size.grid(row=row, column=1)
        row += 1
        ttk.Label(parent, text="Sweep size: ").grid(row=row, column=0, sticky=tk.W)
        self.cl_lbl_sweep_samples = ttk.Label(parent)
        self.cl_lbl_sweep_samples.grid(row=row, column=1)
        return row

    def mainloop(self):
        self.root.mainloop()



from matplotlib.axes import Axes

class Ax:
    def __init__(self, ax: Axes, parent=None):
        self.parent = parent
        self.ax = ax
        self._art = {}
        for i, art in enumerate(ax.artists):
            self.add_artist(i, art)

    @property
    def fig(self):
        return self.ax.figure

    @property
    def canvas(self):
        return self.fig.canvas

    @property
    def artists(self):
        return self._art

    def art(self, name):
        return self._art.get(name, None)

    def add_artist(self, name, art):
        if not art.figure == self.canvas.figure:
            raise RuntimeError
        art.set_animated(True)
        art.set_visible(True)
        self._art[name] = art

    def plot(self, *args, **kwargs):
        name = kwargs.get("name", None)
        if name is not None:
            del kwargs["name"]
        else:
            name = len(self._art)
        if self.art(name) is None:
            line, = self.ax.plot(*args, **kwargs)
            self.add_artist(name, line)
        else:
            line = self.art(name)
            if len(args) == 2:
                line.set_data(*args) # type: ignore
            else:
                print(f"plot args: {len(args)}")
                print(f"plot kwargs: {kwargs}")
        return line

    def plot3d(self, *args, **kwargs):
        name = kwargs.get("name", None)
        if name is not None:
            del kwargs["name"]
        else:
            name = len(self._art)
        if self.art(name) is None:
            line, = self.ax.plot(*args, **kwargs)
            self.add_artist(name, line)
        else:
            line = self.art(name)
            if len(args) == 3:
                line.set_data_3d(*args) # type: ignore
            else:
                print(f"plot args: {len(args)}")
                print(f"plot kwargs: {kwargs}")
        return line

    def scatter(self, *args, **kwargs):
        name = kwargs.get("name", None)
        if name is not None:
            del kwargs["name"]
        else:
            name = len(self._art)
        if self.art(name) is None:
            line, = self.ax.plot(*args, **kwargs)
            self.add_artist(name, line)
        else:
            line = self.art(name)
            if len(args) == 2:
                line.set_data(*args) # type: ignore
            else:
                print(f"plot args: {len(args)}")
                print(f"plot kwargs: {kwargs}")
        return line

    def imshow(self, *args, **kwargs):
        name = kwargs.get("name", None)
        if name is not None:
            del kwargs["name"]
        else:
            name = len(self._art)
        if self.art(name) is None:
            im = self.ax.imshow(*args, **kwargs)
            self.add_artist(name, im)
        else:
            im = self.art(name)
            im.set_data(*args) # type: ignore
            ks = [k for k in kwargs.keys()]
            for k in ks:
                if k in ("cmap", "interpolation", "resample", "rasterized"):
                    im.set(**{k: kwargs[k]}) # type: ignore
                    del kwargs[k]
        return im

    def plot_wireframe(self, *args, **kwargs):
        name = kwargs.get("name", None)
        if name is not None:
            del kwargs["name"]
        else:
            name = len(self._art)
        surf = self.ax.plot_wireframe(*args, **kwargs) # type: ignore
        # self.add_artist(name, surf)
        return surf

    def plot_surface(self, *args, **kwargs):
        name = kwargs.get("name", None)
        if name is not None:
            del kwargs["name"]
        else:
            name = len(self._art)
        surf = self.ax.plot_surface(*args, **kwargs) # type: ignore
        # self.add_artist(name, surf)
        return surf

    def set_data(self, name, *args):
        self.art(name).set_data(*args) # type: ignore

    def set_xdata(self, name, x):
        self.art(name).set_xdata(x) # type: ignore

    def set_ydata(self, name, y):
        self.art(name).set_ydata(y) # type: ignore

    def set_xlim(self, xmin, xmax):
        self.ax.set_xlim(xmin, xmax)
    def set_ylim(self, ymin, ymax):
        self.ax.set_ylim(ymin, ymax)

    def relim(self):
        self.ax.relim()

    def cla(self):
        self.ax.cla()

class BlitAx(Ax):
    def set_xlim(self, *args, **kwargs):
        super().set_xlim(*args, **kwargs)
        self.relim()

    def set_ylim(self, *args, **kwargs):
        super().set_ylim(*args, **kwargs)
        self.relim()

    def relim(self, *args, **kwargs):
        super().relim(*args, **kwargs)
        self.canvas.draw()



import argparse
import tkinter as tk
from tkinter import ttk

import numpy as np

import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

from ..panels import Panel
from ..plot_base import FreqPlotControllerRT

from ....utils import matrix
from ....backend.mpl.color import cmap
from ....utils.window import WindowLUT
from ....backend.mpl.plot import _Plot, Plot, BlitPlot

class PlotConfig:
    ref_level = "auto"
    scale_div = 10.0
    vbw = 0.0
    window = "blackman"
    x = 1001
    y = 600
    cmap = "hot"

def define_args(parser: argparse.ArgumentParser):
    parser.add_argument("-rl", "--ref_level", default=PlotConfig.ref_level, type=float, help="ref Level")
    parser.add_argument("-sd", "--scale_div", default=PlotConfig.scale_div, type=float, help="scale per division")
    parser.add_argument("-vb", "--vbw", default=PlotConfig.vbw, type=float, help="video bandwidth")
    parser.add_argument("-w", "--window", default=PlotConfig.window, choices=[k for k in WindowLUT.keys()], help="FFT window function")

class PST(FreqPlotControllerRT):
    def __init__(self, parent, pane: Panel, **kwargs):
        pane.master.config(text="Persistent")
        self.x = kwargs.get("x", PlotConfig.x)
        self.y = kwargs.get("y", PlotConfig.y)
        self.cmap = kwargs.get("cmap", PlotConfig.cmap)
        self.fmin = None
        self.fmax = None
        if not "ref_level" in kwargs:
            kwargs["ref_level"] = PlotConfig.ref_level
        if not "scale_div" in kwargs:
            kwargs["scale_div"] = PlotConfig.scale_div
        if not "vbw" in kwargs:
            kwargs["vbw"] = PlotConfig.vbw
        if not "window" in kwargs:
            kwargs["vbw"] = PlotConfig.vbw
        super().__init__(parent, pane, **kwargs)
        self._cmap_set = False
        self._cb_drawn = False

        fig = plt.figure(figsize=(5,5), layout="constrained")

        fig.canvas = FigureCanvasTkAgg(fig, master=self.fr_canv)
        self.plotter = BlitPlot(fig) # type: ignore
        self.plotter.canvas.draw()
        self.plotter.canvas.get_tk_widget().pack(side=tk.LEFT, fill=tk.BOTH, expand=True) # type: ignore
        self.plotter.add_ax("pst", fig.add_subplot())

        self.plotter.ax("pst").ax.set_autoscale_on(False)
        self.plotter.ax("pst").ax.locator_params(axis="x", nbins=5)
        self.plotter.ax("pst").ax.locator_params(axis="y", nbins=10)
        self.plotter.ax("pst").ax.grid(True, alpha=0.2)

        self.update_f(self.parent.dispatch.last_f)
        if not self.ref_level == "auto":
            self.set_y()
        self.update()

    def reset(self):
        pass

    def _plot(self, samps):
        psd = self.psd(samps)
        if self.ref_level == "auto":
            self.ref_level = self._calc_ref_level(psd)
            self.set_ref_level(self.ref_level)
            self.set_y()
        self.pane.master.config(text=f"Persistent - {psd.shape[1]} FFTs")

        mat = matrix.cvec(self.x, self.y, psd, self.y_top, self.y_btm)
        mat = mat / np.max(mat)

        im = self.plotter.ax("pst").imshow(
                mat, name="mat", cmap=cmap[self.cmap],
                vmin=0, vmax=1,
                aspect="auto",
                interpolation="nearest", resample=False, rasterized=True
        )

        if not self._cb_drawn:
            # print("Adding colorbar")
            cb = self.plotter.fig.colorbar(
                im, ax=self.plotter.ax("pst").ax, # type: ignore
                pad=0.005, fraction=0.05
            )
            self.plotter.canvas.draw()
            self._cb_drawn = True

        if self._cmap_set:
            self.plotter.ax("pst").set_ylim(0, self.y)
            self._cmap_set = False

        self._show_y_location(psd)
        self.update()

    def update_f(self, f):
        fmin, fmax, fnum = f
        if self.fmin == fmin and self.fmax == fmax:
            return
        else:
            self.fmin = fmin
            self.fmax = fmax
        x_mul = [0.0,0.25,0.5,0.75,1.0]

        x_tick = [self.x*m for m in x_mul]
        x_text = [f"{m-self.x/2:.1f}" for m in x_tick]
        self.plotter.ax("pst").ax.set_xticks(x_tick, x_text)
        self.plotter.ax("pst").set_xlim(0, self.x)
        self.update()

    def update_nfft(self, nfft):
        self.reset()

    def set_y(self):
        """Set plot ylimits"""
        y_mul = [0.0,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1.0]
        y_max = self.y_top
        y_min = self.y_btm
        y_rng = abs(y_max - y_min)
        #y_off = y_min if y_min < 0 else -y_min
        y_off = y_min

        y_tick = [self.y*m for m in y_mul]
        y_text = [f"{(y_rng*m)+y_off:.1f}" for m in y_mul]
        self.plotter.ax("pst").ax.set_yticks(y_tick, y_text)
        self.plotter.ax("pst").set_ylim(0, self.y)

    def draw_settings(self, row=0):
        row = super().draw_settings(row)

        var_cmap = tk.StringVar(self.pane.settings, str(self.cmap))
        cb_cmap = ttk.Combobox(self.pane.settings, textvariable=var_cmap, width=9)
        cb_cmap.configure(values=[k for k in cmap.keys()])
        cb_cmap.bind("<<ComboboxSelected>>", self.handle_event)

        self.pane.wgts["cmap"] = cb_cmap
        self.pane.sets["cmap"] = var_cmap

        ttk.Separator(self.pane.settings, orient=tk.HORIZONTAL).grid(row=row,column=0,columnspan=3, pady=5, sticky=tk.EW)
        row += 1
        ttk.Label(self.pane.settings, text="Colors").grid(row=row, column=0)
        cb_cmap.grid(row=row, column=1)
        row += 1
        return row

    # --- GUI bind events and setters --- #
    def handle_event(self, event):
        if event.widget == self.pane.wgts["cmap"]:
            self.set_cmap(self.pane.sets["cmap"].get())
        else:
            super().handle_event(event)
    def set_scale(self, scale):
        prev = self.scale
        super().set_scale(scale)
        if not prev == self.scale:
            self.set_y()

    def set_ref_level(self, ref):
        prev = self.ref_level
        super().set_ref_level(ref)
        if not prev == self.ref_level:
            self.set_y()

    def set_vbw(self, smooth):
        prev = float(self.vbw)
        super().set_vbw(smooth)
        if not prev == self.vbw:
            pass

    def set_cmap(self, _cmap):
        """Set plot color mapping"""
        self.cmap = _cmap
        self._cmap_set = True



import pysdrlib

from ...utils import args
from .sink import Sink, args_sink, SinkConfig

from ...obj import Frequency

def args_live(parser):
    sink = args.get_group(parser, "Sink (LIVE)")
    args_sink(sink)
    sink.add_argument("-d", "--device", default=None, help="device type")

class SinkLive(Sink):
    __slots__ = (
        "dev", "gain"
    )
    def __init__(self, model, **kwargs):
        self.dev: pysdrlib.Device = None # type: ignore
        device = kwargs.get("device", None)
        super().__init__(model, **kwargs)
        self.gain = {
            "rf": None,
            "if": None,
            "bb": None
        }
        self.set_device(device) # type: ignore

    def show(self, ind=2):
        print(" "*ind + f"{type(self.dev).__name__}")

    @property
    def name(self):
        return f"{type(self.dev).__name__}"
    @property
    def id(self):
        return None

    def set_device(self, name: str):
        self.log.debug("set_device(%s)", name)
        if self.dev is not None:
            if self.dev.state["rx"]:
                self.dev.stop_rx()
            self.dev.close()

        sdr = pysdrlib.devices.get(name)
        if sdr is not None:
            self.dev = sdr.Device()
            self.dev.open()

            cf = None if int(self._cf.raw) == int(SinkConfig.cf) else self._cf.raw
            Fs = None if int(self._Fs.raw) == int(SinkConfig.Fs) else self._Fs.raw
            # self.dev.initialize(cf=cf, Fs=Fs)
            self.dev.set_freq(cf)
            self.dev.set_sample_rate(Fs)
            self._Fs = Frequency.get(self.dev.get_sample_rate())
            self._cf = Frequency.get(self.dev.get_freq())
            self.set_rx(None)
            self.dev.start_rx()

    def _set_fs(self, fs):
        return self.dev.set_sample_rate(fs)
    def _set_cf(self, cf):
        return self.dev.set_freq(cf)

    def next(self, count: int):
        # self.log.debug("next(%s)", count)
        self._samples = self.dev.get_samples()[-count:]
        return True

    def set_rx(self, gain):
        self.log.debug("set_rx_gain(%s)", gain)
        gain = self.dev.set_rx_gain(gain)
        if gain.get("rf", None) is not None:
            self.gain["rf"] = gain["rf"]
        if gain.get("if", None) is not None:
            self.gain["if"] = gain["if"]
        if gain.get("bb", None) is not None:
            self.gain["bb"] = gain["bb"]
    def set_rx_rf(self, gain):
        self.log.debug("set_rx_rf(%s)", gain)
        self.gain["rf"] = self.dev.set_rx_rf_gain(gain)
    def set_rx_if(self, gain):
        self.log.debug("set_rx_if(%s)", gain)
        self.gain["if"] = self.dev.set_rx_if_gain(gain)
    def set_rx_bb(self, gain):
        self.log.debug("set_rx_bb(%s)", gain)
        self.gain["bb"] = self.dev.set_rx_bb_gain(gain)

    def get_rx_rf(self):
        return self.gain["rf"]
    def get_rx_if(self):
        return self.gain["if"]
    def get_rx_bb(self):
        return self.gain["bb"]

    def has_rx_rf(self):
        return self.dev.CONFIG.GAIN_RX_RF
    def has_rx_if(self):
        return self.dev.CONFIG.GAIN_RX_IF
    def has_rx_bb(self):
        return self.dev.CONFIG.GAIN_RX_BB

    def start_rx(self):
        self.log.debug("start_rx()")
        self.dev.start_rx()
    def stop_rx(self):
        self.log.debug("stop_rx()")
        self.dev.stop_rx()
